#############################################################################
# Author: <martin.teichmann@xfel.eu>
# Created in June 2014
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QPoint, Qt, Slot
from qtpy.QtGui import QTextCursor
from qtpy.QtWidgets import QDialog, QMenu, QPlainTextEdit, QSplitter

import karabogui.access as krb_access
from karabo.common.api import ServerFlags
from karabo.common.project.api import write_macro
from karabogui import icons, messagebox
from karabogui.binding.api import PropertyProxy
from karabogui.dialogs.api import DebugRunDialog
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.project.utils import restart_macro, run_macro, run_macro_debug
from karabogui.singletons.api import get_config, get_topology
from karabogui.topology.api import is_macro_online
from karabogui.util import getSaveFileName
from karabogui.widgets.scintilla_editor import CodeBook
from karabogui.widgets.toolbar import ToolBar

from .base import BasePanelWidget


class MacroPanel(BasePanelWidget):

    def __init__(self, model):
        self.model = model
        super().__init__(model.simple_name, allow_closing=True)

        self.already_connected = {}

        # Register to KaraboBroadcastEvent
        self.event_map = {
            KaraboEvent.ConnectMacroInstance: self._event_connect,
            KaraboEvent.DeviceInitReply: self._event_init_reply,
        }
        register_for_broadcasts(self.event_map)

        # Hook up a trait handler for the panel window title
        self.model.on_trait_change(self.set_title, 'simple_name')
        # Connect all running macros, but don't select them
        for instance in model.instances:
            self.connect(instance, select=False)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        widget = QSplitter(Qt.Vertical, parent=self)
        ui_editor = CodeBook(parent=widget, code=write_macro(self.model))
        widget.addWidget(ui_editor)
        ui_editor.codeChanged.connect(self.on_macro_changed)
        ui_editor.qualityChecked.connect(self.updateCodeQualityIcon)
        self.ui_editor = ui_editor

        self.ui_console = QPlainTextEdit(widget)
        self.ui_console.setReadOnly(True)
        self.ui_console.setStyleSheet("font-family: monospace")
        self.ui_console.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui_console.customContextMenuRequested.connect(
            self._show_context_menu)

        widget.addWidget(self.ui_console)

        return widget

    def toolbars(self):
        """This should create and return one or more `ToolBar` instances needed
        by this panel.
        """
        toolbar = ToolBar(parent=self)
        toolbar.addAction(icons.run, "Run", self.on_run)
        toolbar.addAction(icons.logDebug, "Debug Run", self.on_debug_run)
        toolbar.addAction(icons.save, "Save", self.on_save)
        toolbar.addSeparator()
        toolbar.addAction(icons.zoomIn, "Increase font", self.increaseFont)
        toolbar.addAction(icons.zoomOut, "Decrease font", self.decreaseFont)

        # Buttons for linter
        toolbar.addSeparator()
        self.check_button = toolbar.addAction(
            icons.inspect_code, "Check Code Quality", self.checkCode)
        self.clear_button = toolbar.addAction(
            icons.editClear, "Clear Linter Messages", self.clearIndicators)
        self.clear_button.setVisible(False)
        return [toolbar]

    @Slot()
    def checkCode(self):
        """Run linters to check code quality and highlight with indicators"""
        self.ui_editor.checkCode()

    @Slot()
    def clearIndicators(self):
        """Clear the linter indicators"""
        self.ui_editor.clearIndicators()
        self.clear_button.setVisible(False)

    @Slot(bool)
    def updateCodeQualityIcon(self, error):
        """ Set the icon on 'check_button' depending on whether the code
        has any error or not. """
        icon = icons.inspect_code_error if error else icons.inspect_code_ok
        self.check_button.setIcon(icon)
        self._setClearButtonVisibility()

    @Slot()
    def increaseFont(self):
        self.ui_editor.increaseFontSize()

    @Slot()
    def decreaseFont(self):
        self.ui_editor.decreaseFontSize()

    # -----------------------------------------------------------------------

    def setReadOnly(self, value):
        """This method is externally called for `AccessLevel` dependent view"""
        self.ui_editor.setReadOnly(value)

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_connect(self, data):
        model = data.get('model')
        if model is self.model:
            self.connect(data.get('instance'))

    def _event_init_reply(self, data):
        macro_instance = data.get('device')
        if macro_instance.device_id in self.model.instance_id:
            self.init_reply(data.get('success'), data.get('message'))

    # -----------------------------------------------------------------------

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            # Unregister to KaraboBroadcastEvent
            unregister_from_broadcasts(self.event_map)
            # Unregister the trait handler too
            self.model.on_trait_change(self.set_title, 'simple_name',
                                       remove=True)
            # Tell the world we're closing
            broadcast_event(KaraboEvent.MiddlePanelClosed,
                            {'model': self.model})
            # Unregister all trait handlers
            for instance in self.model.instances:
                self.disconnect(instance)

    def connect(self, macro_instance, select=True):
        if macro_instance not in self.already_connected:
            macro_proxy = get_topology().get_device(macro_instance)
            self.ui_console.moveCursor(QTextCursor.End)
            self.ui_console.insertPlainText('Connecting to {} '
                                            '...'.format(macro_instance))
            macro_proxy.on_trait_change(self.finish_connection,
                                        'schema_update')
            if len(macro_proxy.binding.value) != 0:
                # if the proxy is already populated trigger the action
                self.finish_connection(macro_proxy,
                                       name='schema_update', new=None)
        if select:
            broadcast_event(KaraboEvent.ShowProjectDevice,
                            {"deviceId": macro_instance})

    def disconnect(self, macro_instance):
        if macro_instance in self.already_connected:
            text_proxy = self.already_connected.pop(macro_instance)
            text_proxy.stop_monitoring()
            text_proxy.on_trait_change(self.append_console,
                                       'binding.config_update', remove=True)

    def finish_connection(self, macro_proxy, name, new):
        if macro_proxy.device_id not in self.already_connected:
            text_proxy = PropertyProxy(root_proxy=macro_proxy, path='print')
            text_proxy.on_trait_change(
                self.append_console, 'binding.config_update')
            # start monitoring property
            text_proxy.start_monitoring()
            self.ui_console.moveCursor(QTextCursor.End)
            self.ui_console.insertPlainText('Connection done!\n')
            self.already_connected[macro_proxy.device_id] = text_proxy
        macro_proxy.on_trait_change(self.finish_connection,
                                    'schema_update', remove=True)

    def append_console(self, binding, name, new):
        if name != 'config_update':
            return
        self.ui_console.moveCursor(QTextCursor.End)
        self.ui_console.insertPlainText(binding.value)

    def init_reply(self, ok, message):
        self.ui_console.moveCursor(QTextCursor.End)
        self.ui_console.insertPlainText(message)
        self.ui_console.insertPlainText("\n")
        self.ui_console.moveCursor(QTextCursor.End)

    @Slot(QPoint)
    def _show_context_menu(self, pos):
        """Show a context menu"""
        menu = QMenu()
        clear_action = menu.addAction('Clear Console')
        clear_action.triggered.connect(self.ui_console.clear)
        select_action = menu.addAction('Select All')
        select_action.triggered.connect(self.ui_console.selectAll)
        menu.exec(self.ui_console.viewport().mapToGlobal(pos))

    @Slot()
    def on_run(self):
        if not self.has_access_level():
            return

        instance_id = self.model.instance_id
        if is_macro_online(instance_id):
            restart_macro(instance_id, self._run_code)
        else:
            self._run_code()

    def _run_code(self):
        self.ui_console.clear()
        try:
            compile(self.model.code, self.model.simple_name, "exec")
        except SyntaxError as e:
            if e.filename == self.model.simple_name:
                self.ui_editor.moveCursorToLine(e.lineno, e.offset)
            formatted_msg = "{}\n{}{}^\nin {} line {}".format(
                e.msg, e.text, " " * e.offset, e.filename, e.lineno)
            messagebox.show_warning(formatted_msg, title=type(e).__name__,
                                    parent=self)
            return

        run_macro(self.model, parent=self)

    @Slot()
    def on_debug_run(self):
        """
        Show a dialog with MacroServers with "Development" serverflag and
        run the Macro on the selected macro server.
        """
        if not self.has_access_level():
            return

        topology = get_topology()
        servers = []

        def visitor(node):
            attributes = node.attributes
            if attributes.get("lang") == "macro" and (
                    attributes.get("serverFlags", 0)
                    & ServerFlags.Development == ServerFlags.Development):
                servers.append(node.path)

        topology.visit_system_tree(visitor)
        if not servers:
            messagebox.show_error(
                "No (development) Macro server found in system "
                "topology. Macro cannot be started.")
            return

        dialog = DebugRunDialog(serverIds=servers, parent=self)
        if dialog.exec() == QDialog.Accepted:
            server_id = dialog.comboBox.currentText()
            run_macro_debug(self.model, serverId=server_id, parent=self)
            get_config()["macro_development"] = server_id

    @Slot()
    def on_save(self):
        default_file_name = self.model.simple_name.replace("/", "_")
        fn = getSaveFileName(
            caption="Save macro to file",
            filter="Python files (*.py)",
            suffix="py",
            selectFile=default_file_name + ".py")
        if not fn:
            return

        with open(fn, "w") as out:
            out.write(write_macro(self.model))

    @Slot()
    def on_macro_changed(self):
        self.model.code = self.ui_editor.getEditorCode()
        self.check_button.setIcon(icons.inspect_code)
        self._setClearButtonVisibility()

    def _setClearButtonVisibility(self):
        """ The 'clear_button' should be visible only when the issues are
        annotated."""
        has_annotation = self.ui_editor.code_editor.has_annotation
        self.clear_button.setVisible(has_annotation)

    def __repr__(self):
        return f"<MacroPanel macro={self.model.simple_name}>"

    def has_access_level(self):
        role = krb_access.AccessRole.SERVICE_EDIT
        allowed = krb_access.access_role_allowed(role)
        if not allowed:
            required = krb_access.get_access_level_for_role(role)
            msg = ("The current access level "
                   f"'{krb_access.GLOBAL_ACCESS_LEVEL.name}' "
                   "is not sufficient to run the macro! Access level "
                   f"{required} is required.")
            messagebox.show_information(msg, parent=self)
            return False

        return True



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QWidget
from traits.api import Any, Dict, HasStrictTraits, Instance, List

from karabo.common.project.api import ProjectModel
from karabogui.singletons.api import get_config, get_network

from .controllers import (
    ConfigureController, LinkController, SelectScenesController)
from .pages import PageContainer


class CinemaWizardController(HasStrictTraits):
    # The QWizard PageContainer
    widget = Instance(PageContainer)

    # Parent widget, can be None
    parent = Instance(QWidget, transient=True)

    # Shared variables
    project_model = Instance(ProjectModel)
    available_scenes = List
    selected_scenes = List
    link = Dict

    # Pages
    page_controller = List
    current_controller = Any

    def run(self):
        """Run the cinema wizard for the different entry points"""
        if len(self.selected_scenes):
            # Started directly on a scene
            self._set_start_page(ConfigureController)
        elif self.project_model is not None:
            # Started for the scene group
            self._set_start_page(SelectScenesController)
            self.available_scenes = self.project_model.scenes

        self.widget.show()

    # ----------------------------------------------------------------------
    # Trait methods

    def _widget_default(self):
        wizard = PageContainer(parent=self.parent)
        # Add page_controller
        for index, controller in enumerate(self.page_controller):
            wizard.setPage(index, controller.page)
        # Connect signals
        wizard.currentIdChanged.connect(self._set_current_controller)

        return wizard

    def _page_controller_default(self):
        return [
            SelectScenesController(wizard_controller=self),
            ConfigureController(wizard_controller=self),
            LinkController(wizard_controller=self)]

    def _link_default(self):
        network = get_network()
        return dict(
            domain=get_config()["domain"],
            host=network.hostname,
            port=network.port,
            include_host=False,
            show_splash=True,
            uuids=[])

    def _current_controller_changed(self, old, new):
        if old is not None:
            old.exit()
        if new is not None:
            # Check if page is already visited
            if not self._has_visited(old, new):
                new.init()
            new.enter()

    # ----------------------------------------------------------------------
    # Private methods

    def _set_current_controller(self, index):
        self.current_controller = self.page_controller[index]

    def _set_start_page(self, klass):
        for index, page in enumerate(self.page_controller):
            if isinstance(page, klass):
                self.widget.setStartId(index)
                return

        raise IndexError(f"No page_controller of class {klass} found.")

    def _has_visited(self, old, new):
        def get_id(page):
            page_id = -1
            if page is not None:
                page_id = self.page_controller.index(page)
            return page_id

        return get_id(new) < get_id(old)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtGui import QValidator

from ..tool_widget import SearchValidator

LONG_NAME = "middlelayerServer/veryveryveryveryveryverylongnameforaserver" \
            "thatshouldnotbethere"


def test_validator():
    """Test the search validator for the navigation panels"""
    validator = SearchValidator()

    # Empty string must always be `ok`
    valid, *_ = validator.validate("", None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate("aAbBC", None)
    assert valid is QValidator.Acceptable

    # White space does not work
    valid, *_ = validator.validate(" ", None)
    assert valid is QValidator.Invalid

    valid, *_ = validator.validate("aBa ", None)
    assert valid is QValidator.Invalid

    # Normal devices and servers

    valid, *_ = validator.validate("SA4_RR_SYS/PLC/RACK-9", None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate("SA4_RR_SYS/PLC/RACK-9/BROKEN", None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate("cppServer/beckhoff_mc2", None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate("////", None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate("////---//", None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate(LONG_NAME, None)
    assert valid is QValidator.Acceptable

    valid, *_ = validator.validate("#", None)
    assert valid is QValidator.Invalid

    valid, *_ = validator.validate(":", None)
    assert valid is QValidator.Invalid



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import argparse
import sys

from qtpy.QtWidgets import QApplication

from karabo.common.scenemodel.api import (
    DeviceSceneLinkModel, SceneModel, read_scene)
from karabogui import icons
from karabogui.controllers.api import populate_controller_registry
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.panels.configurationpanel import ConfigurationPanel
from karabogui.panels.navigationpanel import TopologyPanel
from karabogui.panels.projectpanel import ProjectPanel
from karabogui.panels.scenepanel import ScenePanel
from karabogui.singletons.api import (
    get_manager, get_mediator, get_network, get_panel_wrangler, get_topology)
from karabogui.util import getOpenFileName


def run_configurator(ns):
    device_id = ns.configurator
    device = get_topology().get_device(device_id)
    broadcast_event(KaraboEvent.ShowConfiguration, {'proxy': device})
    return ConfigurationPanel(), (600, 800)


def run_navigation():
    return TopologyPanel(), (400, 600)


def run_project():
    return ProjectPanel(), (300, 500)


def run_scene():
    filename = getOpenFileName(filter='*.svg')
    if not filename:
        sys.exit()
    return run_scene_file(filename)


def run_scene_file(filename):
    model = read_scene(filename)
    panel = ScenePanel(model, True)

    return panel, (1024, 768)


def run_scene_link(ns):
    device_id, scene_name = ns.dev_scene_link.split('|')
    keys = [f"{device_id}.availableScenes"]
    link = DeviceSceneLinkModel(
        keys=keys, target=scene_name, text=ns.dev_scene_link, width=200)
    model = SceneModel(children=[link])
    panel = ScenePanel(model, True)

    return panel, (1024, 768)


def run_panel(ns):
    app = QApplication(sys.argv)
    # Run the lazy initializers (icons, widget controllers)
    icons.init()
    populate_controller_registry()

    # Init some singletons
    get_mediator()
    get_manager()

    if ns.configurator:
        panel, size = run_configurator(ns)
    elif ns.dev_scene_link:
        panel, size = run_scene_link(ns)
    elif ns.navigation:
        panel, size = run_navigation()
    elif ns.project:
        panel, size = run_project()
    elif ns.scene_chooser:
        panel, size = run_scene()
    elif ns.scene_file:
        panel, size = run_scene_file(
            ns.scene_file)

    # XXX: A hack to keep the toolbar visible
    panel.toolbar.setVisible(True)
    panel.show()
    panel.resize(*size)

    # Init the panel wrangler singleton
    get_panel_wrangler()

    # Connect to the GUI Server
    get_network().connectToServer()
    sys.exit(app.exec())


def main():
    ap = argparse.ArgumentParser(description='GUI Panel Runner')
    ag = ap.add_mutually_exclusive_group(required=True)
    ag.add_argument('-c', '--configurator', type=str, metavar='DEVICE_ID',
                    help='A device ID should be provided')
    ag.add_argument(
        '-l', '--dev_scene_link', type=str, metavar='DEVICE|SCENENAME')
    ag.add_argument('-n', '--navigation', action='store_true')
    ag.add_argument('-p', '--project', action='store_true')
    ag.add_argument('-s', '--scene_chooser', action='store_true')
    ag.add_argument('-S', '--scene_file', type=str, metavar='FILENAME')
    run_panel(ap.parse_args())


if __name__ == '__main__':
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QFrame

from karabo.native import AccessLevel
from karabogui import mainwindow
from karabogui.testing import click_button, singletons


def test_mainwindow(gui_app, mocker, subtests):
    """Small test for verifying that at least the imports are right"""

    network = mocker.Mock()
    manager = mocker.Mock()
    mediator = mocker.Mock()
    krb_access = mocker.patch.object(mainwindow, "krb_access")
    krb_access.HIGHEST_ACCESS_LEVEL = AccessLevel.EXPERT
    krb_access.GLOBAL_ACCESS_LEVEL = AccessLevel.EXPERT
    with singletons(network=network, manager=manager, mediator=mediator):
        logger = mocker.patch("karabogui.logger._logger")
        from karabogui.mainwindow import MainWindow
        mw = MainWindow()
        logger.info.assert_called_with("Started Karabo GUI application ...")
        assert mw is not None

        with subtests.test("User Session Button"):
            krb_access.is_authenticated.return_value = True
            krb_access.SERVER_READ_ONLY = False
            mw.onUpdateAccessLevel()
            assert mw.tbUserSession.isVisible()

            mw.update_server_connection(data=None)
            assert not mw.tbUserSession.isVisible()
            assert not mw.tbUserSession.isChecked()
            assert mw.tbUserSession.toolTip() == "User Session Info"

            krb_access.is_authenticated.return_value = False
            krb_access.SERVER_READ_ONLY = True
            mw.onUpdateAccessLevel()
            assert not mw.tbUserSession.isVisible()

    with subtests.test("Test project conflict event in main window"):
        from karabo.common.project.api import ProjectModel
        from karabogui.singletons.project_model import ProjectViewItemModel

        proj = ProjectModel()
        assert proj.conflict is False
        qt_model = ProjectViewItemModel(parent=mw)
        qt_model.root_model = proj
        with singletons(project_model=qt_model):
            logger = mocker.patch("karabogui.mainwindow.get_logger")
            data = {"uuids": [proj.uuid]}
            mw._event_project_updated(data)
            assert proj.conflict is True
            logger.assert_called()

    with subtests.test("Test the banner notification"):
        assert mw.notification_banner.frameStyle() == QFrame.NoFrame

        data = {"message": "Accelerator goes down in 5 min",
                "trash": "doepianango"}
        mw._event_server_notification(data)
        assert mw.notification_banner.frameStyle() == QFrame.Box
        t = "Accelerator goes down in 5 min"
        assert mw.notification_banner.toPlainText() == t

        data = {"message": "", "trash": "doepianango"}
        mw._event_server_notification(data)
        assert mw.notification_banner.toPlainText() == ""
        assert mw.notification_banner.frameStyle() == QFrame.NoFrame

    with subtests.test("Test the menu bar"):
        menu_bar = mw.menuBar()
        main_menus = [act.text() for act in menu_bar.actions()]
        assert len(main_menus) == 5
        assert main_menus == ["&File", "&Links", "&View", "&Tools", "&Help"]

        help_menu = [action.menu() for action in menu_bar.actions() if
                     action.text() == "&Help"][0]
        assert help_menu
        help_actions = help_menu.actions()
        assert len(help_actions) == 5
        expected = [
            "About", "About Qt", "Tips'N'Tricks", "Check for Updates",
            "Keyboard Shortcuts"]
        assert expected == [action.text() for action in help_actions]

    with subtests.test("Test Access Level"):
        krb_access = mocker.patch.object(mainwindow, "krb_access")
        krb_access.HIGHEST_ACCESS_LEVEL = AccessLevel.EXPERT
        krb_access.GLOBAL_ACCESS_LEVEL = AccessLevel.EXPERT
        access_menu = mw.tbAccessLevel.menu()
        access_levels = [act.text() for act in access_menu.actions()]

        # For User-authenticated login, access level up to
        # 'access.HIGHEST_ACCESS_LEVEL' is exposed
        krb_access.is_authenticated.return_value = True
        mw.onUpdateAccessLevel()
        assert access_levels == [
            'Expert', 'Operator', 'Observer']

        # For Access-level login, access level up to
        # 'access.HIGHEST_ACCESS_LEVEL' is exposed
        krb_access.is_authenticated.return_value = False
        mw.onUpdateAccessLevel()
        access_levels = [act.text() for act in access_menu.actions()]
        assert access_levels == ['Expert', 'Operator', 'Observer']

        # The user session button should not be visible for Observer.
        krb_access.GLOBAL_ACCESS_LEVEL = AccessLevel.OBSERVER
        mw.onUpdateAccessLevel()
        assert not mw.tbUserSession.isVisible()

    with subtests.test("Test gui extension update"):
        mw.enableGuiExtensionUpdate()
        assert mw.update_extensions_action.isVisible()

        update_dialog = mocker.patch("karabogui.mainwindow.UpdateDialog")
        click_button(mw.update_gui_extensions)
        assert update_dialog().open.call_count == 1
        assert not mw.update_extensions_action.isVisible()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import inspect
from unittest import main, mock

from karabo.common.project.api import MacroModel, ProjectModel, write_macro
from karabo.common.scenemodel.api import (
    SceneModel, SceneTargetWindow, write_scene)
from karabo.native import Hash
from karabogui.binding.api import (
    DeviceProxy, PropertyProxy, ProxyStatus, apply_configuration,
    apply_default_configuration, build_binding)
from karabogui.events import KaraboEvent
from karabogui.request import (
    get_macro_from_server, get_scene_from_server, onConfigurationUpdate,
    onSchemaUpdate)
from karabogui.singletons.manager import Manager
from karabogui.testing import GuiTestCase, get_simple_schema, singletons


class TestRequestModule(GuiTestCase):

    def setUp(self):
        schema = get_simple_schema()
        root_binding = build_binding(schema)
        apply_default_configuration(root_binding)
        self.root_proxy = DeviceProxy(binding=root_binding, device_id="marty")
        self.root_proxy.status = ProxyStatus.ONLINE
        self.property_proxy = PropertyProxy(
            root_proxy=self.root_proxy, path="foo")

    def test_config_update(self):
        received = 0

        def handler():
            nonlocal received
            received += 1

        for proxy in (self.root_proxy, self.property_proxy):
            received = 0

            onConfigurationUpdate(proxy, handler)

            config = Hash("foo", True, "bar", "a", "charlie", 0)
            apply_configuration(config, self.root_proxy.binding)
            assert received == 1
            apply_configuration(config, self.root_proxy.binding)
            assert received == 1

            handler = onConfigurationUpdate(proxy, handler, remove=False)
            apply_configuration(config, self.root_proxy.binding)
            assert received == 2
            apply_configuration(config, self.root_proxy.binding)
            assert received == 3

            # Remove manually
            proxy.binding.on_trait_change(handler, "config_update",
                                          remove=True)
            apply_configuration(config, self.root_proxy.binding)
            assert received == 3

            network = mock.Mock()
            with singletons(network=network):
                onConfigurationUpdate(proxy, handler, request=True)
                apply_configuration(config, self.root_proxy.binding)
                assert received == 4
                network.onGetDeviceConfiguration.assert_called_with("marty")

    def test_schema_update(self):
        received = 0

        def handler():
            nonlocal received
            received += 1

        for proxy in (self.root_proxy, self.property_proxy):
            received = 0

            onSchemaUpdate(proxy, handler)

            build_binding(get_simple_schema(),
                          existing=self.root_proxy.binding)
            assert received == 1
            build_binding(get_simple_schema(),
                          existing=self.root_proxy.binding)
            assert received == 1

            handler = onSchemaUpdate(proxy, handler, remove=False)
            build_binding(get_simple_schema(),
                          existing=self.root_proxy.binding)
            assert received == 2

            build_binding(get_simple_schema(),
                          existing=self.root_proxy.binding)
            assert received == 3

            # Remove manually
            self.root_proxy.on_trait_change(handler, "schema_update",
                                            remove=True)
            build_binding(get_simple_schema(),
                          existing=self.root_proxy.binding)
            assert received == 3

            network = mock.Mock()
            with singletons(network=network):
                onSchemaUpdate(proxy, handler, request=True, remove=True)
                build_binding(get_simple_schema(),
                              existing=self.root_proxy.binding)
                network.onGetDeviceSchema.assert_called_with("marty")

    def test_get_macro_from_server(self):
        network = mock.Mock()
        with singletons(network=network):
            manager = Manager()
            with singletons(manager=manager):
                # 1. Success case
                project = ProjectModel(simple_name="MyProject")
                assert len(project.macros) == 0

                token = get_macro_from_server("macro_device",
                                              "new_macro", project)
                network.onExecuteGeneric.assert_called_with(
                    "macro_device", "requestMacro", Hash("name", "new_macro"),
                    token=token)

                macro = MacroModel(simple_name="NewMacro", code="import time")
                payload = Hash("success", True, "data", write_macro(macro))
                reply = Hash("payload", payload)
                request = Hash("token", token)

                path = "karabogui.request.broadcast_event"
                with mock.patch(path) as broadcast:
                    manager.handle_requestGeneric(True, request, reply=reply)
                    broadcast.assert_called_with(KaraboEvent.ShowMacroView,
                                                 {"model": mock.ANY})
                # A new macro has been added
                assert len(project.macros) == 1

                # 2. False case
                token = get_macro_from_server("macro_device",
                                              "new_macro", project)
                network.onExecuteGeneric.assert_called_with(
                    "macro_device", "requestMacro", Hash("name", "new_macro"),
                    token=token)
                payload = Hash("success", False, "data", "")
                reply = Hash("payload", payload)
                request = Hash("token", token)
                path = "karabogui.request.messagebox"
                with mock.patch(path) as mbox:
                    manager.handle_requestGeneric(True, request, reply=reply)
                    mbox.show_warning.assert_called_once()

                # No further macro added
                assert len(project.macros) == 1

    def test_get_scene_from_server(self):
        network = mock.Mock()
        with singletons(network=network):
            manager = Manager()
            with singletons(manager=manager):
                # 1. Success case
                project = ProjectModel(simple_name="MyProject")
                assert len(project.scenes) == 0

                token = get_scene_from_server("scene_device",
                                              "new_scene", project)
                network.onExecuteGeneric.assert_called_with(
                    "scene_device", "requestScene", Hash("name", "new_scene"),
                    token=token)

                scene = SceneModel(simple_name="NewScene")
                payload = Hash("success", True, "data", write_scene(scene))
                reply = Hash("payload", payload)
                request = Hash("token", token, "args", Hash())

                path = "karabogui.request.broadcast_event"
                with mock.patch(path) as broadcast:
                    manager.handle_requestGeneric(True, request, reply=reply)
                    broadcast.assert_called_with(
                        KaraboEvent.ShowSceneView,
                        {"model": mock.ANY,
                         'target_window': SceneTargetWindow.Dialog})
                # A new scene has been added
                assert len(project.scenes) == 1

                # 2. False case - from scene protocol
                token = get_scene_from_server("scene_device",
                                              "new_scene", project)
                network.onExecuteGeneric.assert_called_with(
                    "scene_device", "requestScene", Hash("name", "new_scene"),
                    token=token)
                payload = Hash("success", False, "data", "")
                reply = Hash("payload", payload)
                request = Hash("token", token)
                path = "karabogui.request.messagebox"
                with mock.patch(path) as mbox:
                    manager.handle_requestGeneric(True, request, reply=reply)
                    mbox.show_warning.assert_called_once()

                # No further scene added
                assert len(project.scenes) == 1

                # 3. False case - from failing slot call
                token = get_scene_from_server("scene_device",
                                              "new_scene", project)
                network.onExecuteGeneric.assert_called_with(
                    "scene_device", "requestScene", Hash("name", "new_scene"),
                    token=token)
                payload = Hash("success", False, "data", "")
                reply = "failure reason"
                request = Hash("token", token)
                path = "karabogui.request.messagebox"
                with mock.patch(path) as mbox:
                    manager.handle_requestGeneric(False, request, reply=reply)
                    mbox.show_warning.assert_called_once()

                # No further scene added
                assert len(project.scenes) == 1

                # 4. success case, with position
                assert len(project.scenes) == 1
                token = get_scene_from_server("scene_device",
                                              "extra_scene", project=None,
                                              position=[0, 120])

                network.onExecuteGeneric.assert_called_with(
                    "scene_device", "requestScene", Hash("name", "extra_scene",
                                                         "position", [0, 120]),
                    token=token)

                scene = SceneModel(simple_name="NewScene")
                payload = Hash("success", True, "data", write_scene(scene))
                reply = Hash("payload", payload)
                request = Hash("token", token,
                               "args", Hash("position", [0, 120]))

                path = "karabogui.request.broadcast_event"
                with mock.patch(path) as broadcast:
                    manager.handle_requestGeneric(True, request, reply=reply)
                    broadcast.assert_called_with(
                        KaraboEvent.ShowUnattachedSceneView,
                        {"model": mock.ANY,
                         "target_window": SceneTargetWindow.Dialog,
                         "position": [0, 120]})
                # A new scene has not been added
                assert len(project.scenes) == 1

    def test_signature_scene_from_server(self):
        """If this tests fails, we need to check the db connection db scene"""
        sig = inspect.signature(get_scene_from_server)
        params = sig.parameters
        assert "slot_name" in params
        assert "project" in params
        assert "slot_name" in params
        assert "target_window" in params
        assert "scene_name" in params
        assert len(params) == 6


if __name__ == "__main__":
    main()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 1, 2013
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import json

from qtpy.QtCore import QAbstractItemModel, QMimeData, QModelIndex, Qt

from karabo.common.api import InstanceStatus
from karabogui import icons
from karabogui.events import KaraboEvent, register_for_broadcasts
from karabogui.indicators import get_instance_info_icon
from karabogui.singletons.api import get_topology

from .context import _UpdateContext
from .utils import get_language_icon


class SystemTreeModel(QAbstractItemModel):

    def __init__(self, parent=None):
        super().__init__(parent)
        # Our hierarchy tree
        self.tree = get_topology().system_tree
        self.tree.update_context = _UpdateContext(item_model=self)
        self.tree.on_trait_change(self._status_update, 'status_update')

        # Register to KaraboBroadcastEvent, Note: unregister_from_broadcasts is
        # not necessary
        event_map = {
            KaraboEvent.StartMonitoringDevice: self._event_monitor,
            KaraboEvent.StopMonitoringDevice: self._event_monitor,
        }
        register_for_broadcasts(event_map)

    def supportedDragActions(self):
        return Qt.CopyAction

    def _event_monitor(self, data):
        node_id = data['device_id']
        self._update_device_info(node_id)

    def clear(self):
        self.tree.clear_all()

    def index(self, row, column, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_node = self.tree.root
        else:
            parent_node = parent.internalPointer()
            if parent_node is None:
                return QModelIndex()

        children = parent_node.children
        return self.createIndex(row, column, children[row])

    def parent(self, index):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return QModelIndex()

        child_node = index.internalPointer()
        if child_node is None:
            return QModelIndex()

        parent_node = child_node.parent
        if parent_node is None:
            return QModelIndex()

        if parent_node == self.tree.root:
            return QModelIndex()

        return self.createIndex(parent_node.row(), 0, parent_node)

    def rowCount(self, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel. Counts number of
        Children for a given node in the tree view.
        """
        if parent.column() > 0:
            # parent with column number > 0 don't have children
            return 0

        if not parent.isValid():
            parent_node = self.tree.root
        else:
            parent_node = parent.internalPointer()
            if parent_node is None:
                return 0

        return len(parent_node.children)

    def columnCount(self, parentIndex=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        return 2

    def indexInfo(self, index):
        if not index.isValid():
            return {}

        node = index.internalPointer()
        if node is None:
            return {}

        return node.info()

    def data(self, index, role=Qt.DisplayRole):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return

        node = index.internalPointer()
        if node is None:
            return

        column = index.column()
        hierarchyLevel = node.level

        if column == 0 and role == Qt.DisplayRole:
            return node.node_id
        elif column == 0 and role == Qt.DecorationRole:
            if hierarchyLevel == 3:
                if node.status is InstanceStatus.ERROR:
                    return icons.deviceInstanceError
                if node.monitoring:
                    return icons.deviceMonitored
                else:
                    return icons.deviceInstance
            elif hierarchyLevel == 2:
                return icons.deviceClass
            elif hierarchyLevel == 1:
                return icons.yes
            elif hierarchyLevel == 0:
                return icons.host
        elif column == 1 and role == Qt.DecorationRole:
            if hierarchyLevel == 3:
                return get_instance_info_icon(node.status)
            elif hierarchyLevel == 1:
                return get_language_icon(node)

    def flags(self, index):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return Qt.NoItemFlags

        node = index.internalPointer()
        if node is None:
            return Qt.NoItemFlags

        ret = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if node.level == 3:
            # Devices can be dragged into other widgets
            ret |= Qt.ItemIsDragEnabled
            # Devices never have children optimization
            ret |= Qt.ItemNeverHasChildren
        return ret

    def headerData(self, section, orientation, role):
        """Reimplemented function of QAbstractItemModel.
        """
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal and section == 0:
                return "Host - Server - Class - Device"

    def mimeData(self, indices):
        """Reimplemented function of QAbstractItemModel.

        Provide data for Drag & Drop operations.
        """
        # Get one selection per row
        rows = {idx.row(): idx for idx in indices if idx.isValid()}
        # Extract info() dictionaries from SystemTreeNode instances
        data = []
        for idx in rows.values():
            n = idx.internalPointer()
            if n is None:
                continue
            data.append(n.info())

        mimeData = QMimeData()
        mimeData.setData('treeItems', bytearray(json.dumps(data),
                                                encoding='UTF-8'))
        return mimeData

    def _status_update(self, node_ids):
        """Triggered from the status_update Event from the system tree"""
        assert isinstance(node_ids, set)

        for node_id in node_ids:
            node = self.tree.get_instance_node(node_id)
            if node is not None:
                index = self.createIndex(node.row(), 0, node)
                self.dataChanged.emit(index, index, [Qt.DecorationRole])
                index = index.sibling(node.row(), 1)
                self.dataChanged.emit(index, index, [Qt.DecorationRole])

    def _update_device_info(self, node_id, column=0):
        """This function is used to launch a dataChanged signal for a specific
           device Id
        """
        node = self.tree.get_instance_node(node_id)
        if node is not None:
            index = self.createIndex(node.row(), column, node)
            self.dataChanged.emit(index, index, [Qt.DecorationRole])



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtWidgets import QStyledItemDelegate

from karabo.common.scenemodel.api import TableElementModel
from karabo.native import (
    AccessMode, Bool, Configurable, Double, Hash, Int32, String, VectorDouble,
    VectorHash)
from karabogui.binding.config import apply_configuration
from karabogui.controllers.table.api import (
    BaseTableController, BoolButtonDelegate, ColorBindingDelegate,
    ColorNumberDelegate, ProgressBarDelegate, VectorButtonDelegate)
from karabogui.testing import get_property_proxy


class TableSchema(Configurable):
    boolButton = Bool(
        defaultValue=False,
        displayType="TableBoolButton",
        displayedName="Bool Button",
        accessMode=AccessMode.READONLY)
    colorText = String(
        displayType="TableColor|default=red&ginger=blue&bailey=green",
        displayedName="Color Text",
        defaultValue="Indiana",
        accessMode=AccessMode.READONLY)
    colorNumber = Int32(
        defaultValue=1,
        displayType="TableColor|1=blue&2=red&default=black")
    progress = Double(
        displayedName="Progress",
        displayType="TableProgressBar",
        defaultValue=0.0,
        minInc=0.0,
        maxInc=100.0)
    vector = VectorDouble(
        displayedName="VectorDouble",
        displayType="TableVector",
        defaultValue=[])


class Object(Configurable):
    prop = VectorHash(rows=TableSchema)


class KaraboDelegate(QStyledItemDelegate):
    """A delegate class for testing"""


class TableDelegateController(BaseTableController):
    def create_delegates(self):
        delegate = KaraboDelegate(parent=self.tableWidget())
        self.setTableDelegates({0: delegate})


def test_table_delegate_model_view(gui_app):
    """Test the table controller with delegates"""
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = TableElementModel()
    controller = BaseTableController(proxy=proxy,
                                     model=model)
    controller.create(None)
    assert controller.isReadOnly()
    controller.set_read_only(False)
    table_hash = Hash(
        "prop",
        [Hash("boolButton", True, "colorText", "ginger",
              "colorNumber", 10, "progress", 75.0,
              "vector", [1.2, 1.3])])
    apply_configuration(table_hash, proxy.root_proxy.binding)

    widget = controller.tableWidget()
    delegate = widget.itemDelegateForColumn(0)
    assert not isinstance(delegate, BoolButtonDelegate)
    delegate = widget.itemDelegateForColumn(1)
    assert isinstance(delegate, ColorBindingDelegate)
    delegate = widget.itemDelegateForColumn(2)
    assert isinstance(delegate, ColorNumberDelegate)
    delegate = widget.itemDelegateForColumn(3)
    assert isinstance(delegate, ProgressBarDelegate)
    delegate = widget.itemDelegateForColumn(4)
    assert isinstance(delegate, VectorButtonDelegate)

    controller.set_read_only(True)
    delegate = widget.itemDelegateForColumn(0)
    assert isinstance(delegate, BoolButtonDelegate)
    delegate = widget.itemDelegateForColumn(4)
    assert not isinstance(delegate, VectorButtonDelegate)


def test_custom_delegates_controller(gui_app):
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = TableElementModel()
    controller = TableDelegateController(proxy=proxy,
                                         model=model)
    controller.create(None)
    controller.set_read_only(False)
    widget = controller.tableWidget()
    # First column changed
    delegate = widget.itemDelegateForColumn(0)
    assert isinstance(delegate, KaraboDelegate)
    # Others remain have been created as expected
    delegate = widget.itemDelegateForColumn(1)
    assert isinstance(delegate, ColorBindingDelegate)
    delegate = widget.itemDelegateForColumn(2)
    assert isinstance(delegate, ColorNumberDelegate)
    delegate = widget.itemDelegateForColumn(3)
    assert isinstance(delegate, ProgressBarDelegate)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.common.scenemodel.api import SparklineModel
from karabo.native import Configurable, Float
from karabogui.testing import (
    get_class_property_proxy, get_property_proxy, set_proxy_value, singletons)
from karabogui.util import process_qt_events

from ..sparkline import DisplaySparkline
from .data import build_historic_data_float

TIMEBASES = {"60s": 60, "10m": 600, "10h": 36000}


class Object(Configurable):
    prop = Float(defaultValue=0.0)


@pytest.fixture
def sparkline_setup(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    model = SparklineModel()
    controller = DisplaySparkline(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget is not None

    yield controller, proxy

    controller.destroy()
    assert controller.widget is None


def test_basics(sparkline_setup, mocker):
    controller, proxy = sparkline_setup

    # Make sure there is a value with timestamp on it
    set_proxy_value(proxy, "prop", 12.0)

    mocker.patch.object(controller.line_edit, "setVisible")
    sym = "karabogui.controllers.display.sparkline.QInputDialog"
    QInputDialog = mocker.patch(sym)
    for ac in controller.widget.actions():
        ac_name = ac.text()
        if ac_name == "Set value format":
            QInputDialog.getText.return_value = ".3e", True
            ac.trigger()
            assert controller.model.show_format == ".3e"
        elif ac_name == "Show value":
            assert controller.model.show_value == ac.isChecked()
            ac.setChecked(not ac.isChecked())
            assert controller.model.show_value == ac.isChecked()
        elif ac_name == "Use alarm range":
            assert controller.model.alarm_range == ac.isChecked()
            ac.setChecked(not ac.isChecked())
            assert controller.model.alarm_range == ac.isChecked()
        else:
            tb_val = TIMEBASES.get(ac_name[:3], None)
            if tb_val is None:
                continue
            ac.trigger()
            assert tb_val == controller.model.time_base


def test_device(gui_app, mocker):
    network = mocker.Mock()
    with singletons(network=network):
        schema = Object.getClassSchema()
        proxy = get_property_proxy(schema, "prop")

        model = SparklineModel()
        controller = DisplaySparkline(proxy=proxy, model=model)
        controller.create(None)

        proxy.start_monitoring()
        # can't guess the time argument because it's now
        network.onGetPropertyHistory.assert_called()


def test_historic_data(sparkline_setup, mocker):
    controller, proxy = sparkline_setup

    set_proxy_value(proxy, "prop", 42.0)

    historic_data = build_historic_data_float()
    proxy.binding.historic_data = historic_data
    # mock the QPainter, so the code path can be walked deeper down
    # without the need of creating a backend draw device
    mocker.patch("karabogui.controllers.display.sparkline.QPainter")
    controller.render_area.paintEvent(None)
    process_qt_events()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on July 30, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from traits.api import Instance

from karabo.common.scenemodel.api import (
    DisplayAlarmFloatModel, DisplayAlarmIntegerModel, DisplayFloatModel,
    DisplayLabelModel)
from karabogui.binding.api import (
    CharBinding, FloatBinding, IntBinding, StringBinding, get_dtype_format)
from karabogui.controllers.api import (
    AlarmMixin, BaseLabelController, FormatMixin, register_binding_controller)

BINDING_TYPES = (StringBinding, CharBinding, IntBinding, FloatBinding)


@register_binding_controller(ui_name="Value Field",
                             klassname="DisplayLabel",
                             binding_type=BINDING_TYPES, priority=20)
class DisplayLabel(BaseLabelController):
    model = Instance(DisplayLabelModel, args=())

    def binding_update_proxy(self, proxy):
        self.fmt = get_dtype_format(proxy.binding)


@register_binding_controller(ui_name="Float Field",
                             klassname="DisplayFloat",
                             binding_type=FloatBinding,
                             priority=10)
class DisplayFloat(FormatMixin, BaseLabelController):
    model = Instance(DisplayFloatModel, args=())


@register_binding_controller(ui_name="Alarm Float Field",
                             klassname="DisplayAlarmFloat",
                             binding_type=FloatBinding,
                             priority=0)
class DisplayAlarmFloat(AlarmMixin, FormatMixin, BaseLabelController):
    model = Instance(DisplayAlarmFloatModel, args=())


@register_binding_controller(ui_name="Alarm Integer Field",
                             klassname="DisplayAlarmInteger",
                             binding_type=IntBinding,
                             priority=0)
class DisplayAlarmInteger(AlarmMixin, BaseLabelController):
    model = Instance(DisplayAlarmIntegerModel, args=())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtCore import QPoint, QRect
from qtpy.QtWidgets import QWidget

import karabogui.const as global_constants
from karabo.common.api import InstanceStatus
from karabo.common.scenemodel.api import InstanceStatusModel
from karabogui.binding.api import DeviceProxy, ProxyStatus
from karabogui.events import KaraboEvent
from karabogui.testing import check_renderer_against_svg, click_button
from karabogui.topology.api import SystemTreeNode

from ..status import (
    SVG_ERROR, SVG_OFFLINE, SVG_OK, SVG_UNKNOWN, InstanceStatusWidget)


def test_instance_info_widget(gui_app, mocker):
    path = "karabogui.sceneview.widget.status.get_topology"
    topo = mocker.Mock()
    mocker.patch(path).return_value = topo
    root_proxy = DeviceProxy(device_id="XHQ_EH/MDL/STATUS")
    node = SystemTreeNode()
    node.attributes = {"type": "device", "status": InstanceStatus.OK.value}
    root_proxy.status = ProxyStatus.OFFLINE
    topo.get_device.return_value = root_proxy

    model = InstanceStatusModel(
        x=0, y=0, width=100, height=100,
        keys=["XHQ_EH/MDL/STATUS.deviceId"])

    main_widget = QWidget()
    widget = InstanceStatusWidget(model=model,
                                  parent=main_widget)
    main_widget.show()

    model_rect = QRect(model.x, model.y,
                       model.width, model.height)
    assert widget.rect() == model_rect
    assert len(widget.actions()) == 0

    # Geometry
    rect = QRect(2, 10, 2, 10)
    assert widget.geometry() != rect
    widget.set_geometry(rect)
    assert widget.geometry() == rect

    # Translation
    assert widget.pos() == QPoint(2, 10)
    widget.translate(QPoint(10, 0))
    assert widget.pos() == QPoint(12, 10)

    def assert_widget(text, filename):
        assert widget.getToolTip() == f"XHQ_EH/MDL/STATUS - {text}"
        assert widget.renderer().isValid()
        check_renderer_against_svg(widget.renderer(), filename)

    assert not widget.proxy.online
    assert_widget("OFFLINE", SVG_OFFLINE)

    root_proxy.topology_node = node
    node.status = InstanceStatus.OK
    root_proxy.status = ProxyStatus.ONLINE
    assert widget.proxy.online
    assert_widget("OK", SVG_OK)

    node.status = InstanceStatus.OK
    assert_widget("OK", SVG_OK)
    node.status = InstanceStatus.ERROR
    assert_widget("ERROR", SVG_ERROR)
    node.status = InstanceStatus.UNKNOWN
    assert_widget("UNKNOWN", SVG_UNKNOWN)
    node.status = InstanceStatus.OK
    assert_widget("OK", SVG_OK)
    # offline again
    root_proxy.status = ProxyStatus.OFFLINE
    assert_widget("OFFLINE", SVG_OFFLINE)

    path = "karabogui.sceneview.widget.status.broadcast_event"
    broadcast = mocker.patch(path)

    click_button(widget)
    if not global_constants.APPLICATION_MODE:
        broadcast.assert_called_with(KaraboEvent.ShowDevice,
                                     {"deviceId": "XHQ_EH/MDL/STATUS"})
    else:
        broadcast.assert_called_with(
            KaraboEvent.RaiseEditor, {"proxy": root_proxy})

    # Make sure, basic interface is there
    widget.add_proxies(None)
    widget.apply_changes()
    widget.decline_changes()
    widget.set_visible(True)
    widget.update_global_access_level(None)
    widget.destroy()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 9, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox

from karabogui.singletons.api import get_topology
from karabogui.util import InputValidator

from .utils import get_dialog_ui


class DeviceHandleDialog(QDialog):
    def __init__(self, server_id=None, model=None, add_config=False,
                 is_online=False, class_id='', parent=None):
        """ A dialog to configure device configurations

        :param server_id: The ID of the server the device configuration belongs
                          to
        :param model: The ``DeviceInstanceModel`` object
        :param add_config: A boolean which describes whether a new
                           ``DeviceConfigurationModel`` should be added
        :param is_online: A boolean which is True if the device being edited
                          is currently online.
        :param class_id: A string containing the class_id of the class which
                         must be used by the device.
        :param parent: The parent of the dialog
        """
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('device_handle.ui'), self)
        validator = InputValidator(parent=self)
        self.leTitle.setValidator(validator)

        # Get available plugins from systemTopology
        for cls_id in self._get_available_plugins(server_id):
            self.cbClass.addItem(cls_id)
        self.leServerId.setText(server_id)

        # Disable the instance_id editor when the device is online
        self.leTitle.setEnabled(not is_online)

        if model is None:
            title = 'Add device'
            self.cbConfig.setEditable(True)
            self.cbConfig.lineEdit().setText('default')
            # We do not allow to modify 'default' configuration!
            self.cbConfig.setEnabled(False)

            # If we already know the class, select it. It is offline
            if class_id != '':
                self._update_plugin_widget(class_id)
                self.cbClass.setEnabled(False)
        else:
            active_dev_conf = model.select_config(model.active_config_ref)
            if add_config:
                title = 'Add device configuration'
                self.cbConfig.setEditable(True)
                line_edit = self.cbConfig.lineEdit()
                line_edit.setValidator(InputValidator(parent=self))
                line_edit.setFocus()

                # These widgets belong to a ``DeviceInstanceModel`` and
                # should not be changed in case a configuration is added
                self.leTitle.setEnabled(False)
                if active_dev_conf is not None:
                    self._update_plugin_widget(active_dev_conf.class_id)
                self.cbClass.setEnabled(False)

            else:
                title = 'Edit device configuration'
                self._init_config_widgets(model)
                self.cbConfig.setEnabled(not is_online)
                if active_dev_conf is not None:
                    self._update_config_widgets(active_dev_conf)
                    index = self.cbConfig.findText(active_dev_conf.simple_name)
                    self.cbConfig.setCurrentIndex(index)

                self.cbClass.setEnabled(not is_online)

            self.leTitle.setText(model.instance_id)

        self.setWindowTitle(title)
        self.leTitle.textChanged.connect(self._update_button_box)
        self.cbConfig.currentIndexChanged.connect(self._update_button_box)
        self.cbConfig.editTextChanged.connect(self._update_button_box)
        self.cbClass.currentIndexChanged.connect(self._update_button_box)
        self._update_button_box()

    def _init_config_widgets(self, dev_inst_model):
        """ Init all device configuration related widgets to the associated
        ``dev_inst_model``

        :param dev_inst_model: The ``DeviceInstanceModel`` object
        """
        for config in dev_inst_model.configs:
            # Add simple_name to combobox and add uuid of config
            self.cbConfig.addItem(config.simple_name, config)

        # Make sure the signal is triggered when setting the index below
        self.cbConfig.setCurrentIndex(-1)
        # Update if configuration changes
        self.cbConfig.currentIndexChanged[int].connect(self.config_changed)

    def _get_available_plugins(self, device_server_id):
        """Get all available plugins of `systemTopology` for the given
        ``device_server_id``
        """
        attrs = get_topology().get_attributes(f"server.{device_server_id}")
        attrs = attrs if attrs is not None else {}
        available_plugins = attrs.get("deviceClasses", [])
        return sorted(available_plugins)

    def _update_config_widgets(self, dev_config_model):
        """ Update all relevant widgets which show `DeviceConfigurationModel`
        information

        :param dev_config_model: The `DeviceConfgurationModel` which data
                                 should be displayed.
        """
        self._update_plugin_widget(dev_config_model.class_id)

    def _update_plugin_widget(self, class_id):
        index = self.cbClass.findText(class_id)
        self.cbClass.setCurrentIndex(index)

    @Slot()
    def _update_button_box(self):
        """Only enable Ok button, if title and configuration is set
        """
        enabled = (len(self.leTitle.text()) and
                   len(self.cbConfig.currentText()) and
                   len(self.cbClass.currentText()))
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    @Slot(int)
    def config_changed(self, index):
        config_model = self.cbConfig.itemData(index)
        # Update dialog to active configuration data
        self._update_config_widgets(config_model)

    @property
    def instance_id(self):
        return self.leTitle.text()

    @property
    def class_id(self):
        return self.cbClass.currentText()

    @property
    def active_uuid(self):
        index = self.cbConfig.currentIndex()
        config_model = self.cbConfig.itemData(index)
        return config_model.uuid

    @property
    def configuration_name(self):
        return self.cbConfig.currentText()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main

from qtpy.QtWidgets import QWizard

from karabogui.singletons.configuration import Configuration
from karabogui.testing import GuiTestCase, singletons

from ..wizard import TipsTricksWizard


class TestTipWizard(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.wizard = TipsTricksWizard()

    def tearDown(self):
        super().tearDown()
        self.wizard = None

    def test_basic_wizard(self):
        assert len(self.wizard.pageIds()) == 8
        assert self.wizard.windowTitle() == "Karabo Tips & Tricks"
        assert self.wizard.wizardStyle() == QWizard.ClassicStyle

        assert self.wizard.startId() == 0

        assert self.wizard.currentId() == -1
        self.wizard.restart()
        assert self.wizard.currentId() == 0
        self.wizard.next()
        assert self.wizard.currentId() == 1

        # Test the deactivation for next time
        config = Configuration()
        with singletons(configuration=config):
            self.wizard.update_start(0)
            assert config["wizard"] is True
            self.wizard.update_start(1)
            assert config["wizard"] is False

    def test_button_object_names(self):
        next_button = self.wizard.button(QWizard.NextButton)
        assert next_button.objectName() == "NextButton"
        back_button = self.wizard.button(QWizard.BackButton)
        assert back_button.objectName() == "BackButton"


if __name__ == "__main__":
    main()



from qtpy import uic
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtGui import QColor, QTextFormat
from qtpy.QtWidgets import QDialog, QTextEdit

from .utils import get_dialog_ui

HIGHLIGHT_COLOR = QColor(Qt.yellow).lighter(180)
MIN_SIZE = 0
MAX_SIZE = 16777215


class PopupButtonDialog(QDialog):
    """The dialog provides a text editor. The size of the editor is stored
    in the model."""
    def __init__(self, model, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("popup_button_dialog.ui"), self)
        self.model = model.clone_traits()
        self.update_plain_widget()

    def update_plain_widget(self):
        model = self.model
        self.leText.setPlainText(model.text)
        self.leText.setFixedSize(
            QSize(int(model.popup_width), int(model.popup_height)))

    def resizeEvent(self, event):
        """Store the size of the QPlainTextEdit in the model when dialog is
        resized. The popup widget will have the same size"""
        super().resizeEvent(event)
        if self.model.popup_width != self.leText.width():
            self.model.popup_width = self.leText.width()
        if self.model.popup_height != self.leText.height():
            self.model.popup_height = self.leText.height()

    def _make_text_edit_resizable(self):
        """Allow the QPlainTextEdit resizable"""
        min_size = QSize(MIN_SIZE, MIN_SIZE)
        max_size = QSize(MAX_SIZE, MAX_SIZE)
        self.leText.setMinimumSize(min_size)
        self.leText.setMaximumSize(max_size)

    def showEvent(self, event):
        # Before showing the dialog remove the fixed size constraint from
        # the leText QPlainTextEdit.
        self._make_text_edit_resizable()
        super().showEvent(event)

    # ----------------------------------------------------------------------
    # Qt Slots

    @Slot()
    def on_leText_cursorPositionChanged(self):
        extra_selections = []
        selection = QTextEdit.ExtraSelection()
        selection.format.setBackground(HIGHLIGHT_COLOR)
        selection.format.setProperty(QTextFormat.FullWidthSelection, True)
        selection.cursor = self.leText.textCursor()
        selection.cursor.clearSelection()
        extra_selections.append(selection)
        self.leText.setExtraSelections(extra_selections)

    @Slot()
    def on_leText_textChanged(self):
        self.model.text = self.leText.toPlainText()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on May 20, 2014
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QLabel

from .base import BasePanelWidget


class PlaceholderPanel(BasePanelWidget):
    def __init__(self):
        super().__init__("Start Page")
        # Logbook is not needed for this panel.
        self.application_toolbar.setVisible(False)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        return QLabel(self)

    def attach_to_container(self, container):
        """We override this method to make sure the standard toolbar stays
        hidden for this panel.
        """
        super().attach_to_container(container)
        self.standard_toolbar.setVisible(False)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from datetime import UTC, datetime
from platform import system

import pytest

from karabo.native import Configurable, String
from karabogui.controllers.trendmodel import UPTIME
from karabogui.testing import get_class_property_proxy, set_proxy_value
from karabogui.util import process_qt_events

from ..time_graphs import DisplayStateGraph
from .data import build_historic_state_string


class Object(Configurable):
    state = String(defaultValue="ON")


@pytest.fixture
def display_state_graph_setup(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "state")
    controller = DisplayStateGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    yield controller, proxy

    controller.destroy()
    assert controller.widget is None


def test_set_value(display_state_graph_setup):
    controller, proxy = display_state_graph_setup
    set_proxy_value(proxy, "state", "ACQUIRING")
    process_qt_events()


def test_initial_state(display_state_graph_setup):
    controller, _ = display_state_graph_setup
    assert controller._x_detail == UPTIME


def test_actions(display_state_graph_setup):
    controller, _ = display_state_graph_setup
    actions = controller.widget.actions()
    texts = [action.text() for action in actions]
    assert len(actions) == 7
    assert "Range Y" not in texts
    assert "View" in texts


def test_historic_data(display_state_graph_setup):
    controller, proxy = display_state_graph_setup
    data = build_historic_state_string()
    t0 = data[0]["v", ...]["sec"]
    t1 = data[-1]["v", ...]["sec"]

    controller.set_time_interval(t0, t1)
    proxy.binding.historic_data = data
    process_qt_events()


@pytest.mark.skip(reason="The real range is being disturbed in CI")
def test_datetime_update(display_state_graph_setup):
    controller, _ = display_state_graph_setup
    view_box = controller._plot.plotItem.vb
    view_box.setXRange(1570536895, 1571536895, padding=0.05)

    dt_start = controller.widget.dt_start.dateTime()
    dt_end = controller.widget.dt_end.dateTime()

    assert dt_start.toString() == "Tue Oct 8 00:21:35 2019"
    assert dt_end.toString() == "Sun Oct 20 17:54:55 2019"


@pytest.mark.skipif(system() == "Windows",
                    reason="This tests is Unix specific")
def test_range_update(display_state_graph_setup):
    controller, _ = display_state_graph_setup
    plot_view = controller._plot
    plot_item = plot_view.plotItem

    view_box = plot_item.vb
    view_box.setXRange(datetime(1378, 12, 31).timestamp(),
                       datetime(2070, 12, 31).timestamp(), update=False)

    x_min, x_max = plot_item.getAxis("bottom").range
    assert datetime.fromtimestamp(x_min, UTC).year == 1970
    assert datetime.fromtimestamp(x_max, UTC).year == 2038



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest
from qtpy.QtCore import Qt

from karabo.common.scenemodel.api import (
    DoubleLineEditModel, EditableRegexModel, IntLineEditModel)
from karabo.common.states import State
from karabo.native import (
    Configurable, Double, Float, Int32, RegexString, String, UInt8)
from karabogui.binding.api import build_binding
from karabogui.binding.util import get_editor_value
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..lineedit import DoubleLineEdit, EditRegex, Hexadecimal, IntLineEdit


class IntObject(Configurable):
    state = String(defaultValue=State.INIT)
    prop = Int32(minInc=-10, maxInc=10,
                 allowedStates=[State.INIT])


class DoubleObject(Configurable):
    state = String(defaultValue=State.INIT)
    prop = Double(minInc=-1000, maxInc=1000,
                  allowedStates=[State.INIT])


class FloatObject(Configurable):
    state = String(defaultValue=State.INIT)
    prop = Float(defaultValue=0.0001)


class Uint8Object(Configurable):
    prop = UInt8()


class StringObject(Configurable):
    prop = RegexString(regex="^(0|1|[T]rue|[F]alse)")


class RegexEditSetUp:
    def __init__(self):
        self.proxy = get_class_property_proxy(StringObject.getClassSchema(),
                                              "prop")
        self.controller = EditRegex(proxy=self.proxy,
                                    model=EditableRegexModel())
        self.controller.create(None)
        self.controller.set_read_only(False)
        build_binding(StringObject.getClassSchema(),
                      existing=self.proxy.root_proxy.binding)
        self.controller.binding_update(self.proxy)
        self.normal_sheet = self.controller._style_sheet.format("black")
        self.error_sheet = self.controller._style_sheet.format("red")

    @property
    def sheet(self):
        return self.controller.internal_widget.styleSheet()

    @property
    def text(self):
        return self.controller.internal_widget.text()

    @text.setter
    def text(self, value):
        return self.controller.internal_widget.setText(value)


@pytest.fixture
def regex_edit_setup(gui_app):
    test_object = RegexEditSetUp()
    yield test_object
    # teardown
    test_object.controller.destroy()
    assert test_object.controller.widget is None


def test_regex_lineedit_focus_policy(regex_edit_setup):
    regex_lineedit = regex_edit_setup
    assert (regex_lineedit.controller.internal_widget.focusPolicy()
            == Qt.StrongFocus)


def test_regex_lineedit_set_value(regex_edit_setup):
    regex_lineedit = regex_edit_setup
    set_proxy_value(regex_lineedit.proxy, "prop", "1")
    assert regex_lineedit.text == "1"

    set_proxy_value(regex_lineedit.proxy, "prop", "2")
    assert regex_lineedit.text == "2"

    set_proxy_value(regex_lineedit.proxy, "prop", "False")
    assert regex_lineedit.text == "False"


def test_regex_lineedit_edit_value(regex_edit_setup):
    regex_lineedit = regex_edit_setup
    regex_lineedit.text = "2"
    assert regex_lineedit.proxy.edit_value is None
    regex_lineedit.text = "1"
    assert regex_lineedit.proxy.edit_value is not None


def test_regex_lineedit_decline_color(regex_edit_setup):
    regex_lineedit = regex_edit_setup
    regex_lineedit.text = "2.0"
    assert regex_lineedit.proxy.edit_value is None
    assert regex_lineedit.sheet == regex_lineedit.error_sheet
    regex_lineedit.controller.on_decline()
    assert regex_lineedit.sheet == regex_lineedit.normal_sheet


class NumberLineEditSetUp:
    def __init__(self):
        self.double_proxy = get_class_property_proxy(
            DoubleObject.getClassSchema(), "prop")
        self.double_controller = DoubleLineEdit(proxy=self.double_proxy,
                                                model=DoubleLineEditModel())
        self.double_controller.create(None)
        self.double_controller.set_read_only(False)

        self.float_proxy = get_class_property_proxy(
            FloatObject.getClassSchema(), "prop")
        self.float_controller = DoubleLineEdit(proxy=self.float_proxy,
                                               model=DoubleLineEditModel())
        self.float_controller.create(None)
        self.float_controller.set_read_only(False)

        self.int_proxy = get_class_property_proxy(
            IntObject.getClassSchema(), "prop")
        self.int_controller = IntLineEdit(proxy=self.int_proxy,
                                          model=IntLineEditModel())
        self.int_controller.create(None)
        self.int_controller.set_read_only(False)

    @property
    def float_text(self):
        return self.float_controller.internal_widget.text()

    @float_text.setter
    def float_text(self, text):
        self.float_controller.internal_widget.setText(text)

    @property
    def double_text(self):
        return self.double_controller.internal_widget.text()

    @double_text.setter
    def double_text(self, text):
        self.double_controller.internal_widget.setText(text)

    @property
    def int_text(self):
        return self.int_controller.internal_widget.text()

    @int_text.setter
    def int_text(self, text):
        self.int_controller.internal_widget.setText(text)


@pytest.fixture
def number_lineedit_setup(gui_app):
    test_object = NumberLineEditSetUp()
    yield test_object
    # teardown
    test_object.double_controller.destroy()
    assert test_object.double_controller.widget is None
    test_object.float_controller.destroy()
    assert test_object.float_controller.widget is None
    test_object.int_controller.destroy()
    assert test_object.int_controller.widget is None


def test_number_lineedit_set_value(number_lineedit_setup):
    number_lineedit = number_lineedit_setup
    set_proxy_value(number_lineedit.double_proxy, "prop", np.float32(0.00123))
    assert number_lineedit.double_text == "0.00123"
    set_proxy_value(number_lineedit.double_proxy, "prop", 5.4)
    assert number_lineedit.double_text == "5.4"
    set_proxy_value(number_lineedit.double_proxy, "prop", np.float64(0.0088))
    assert number_lineedit.double_text == "0.0088"

    set_proxy_value(number_lineedit.float_proxy, "prop", np.float32(0.00123))
    assert number_lineedit.float_text == "0.00123"
    set_proxy_value(number_lineedit.float_proxy, "prop", np.float64(0.0088))
    assert number_lineedit.float_text == "0.0088"


def test_number_lineedit_state_update(number_lineedit_setup):
    number_lineedit = number_lineedit_setup
    set_proxy_value(number_lineedit.double_proxy, "state", "CHANGING")
    assert not number_lineedit.double_controller.internal_widget.isEnabled()
    set_proxy_value(number_lineedit.double_proxy, "state", "INIT")
    assert number_lineedit.double_controller.internal_widget.isEnabled()


def test_number_lineedit_edit_value(number_lineedit_setup):
    number_lineedit = number_lineedit_setup
    number_lineedit.double_text = "3.14"
    assert abs(number_lineedit.double_proxy.edit_value - 3.14) < 0.0001
    number_lineedit.double_text = "3.14e-2"
    assert abs(number_lineedit.double_proxy.edit_value - 0.0314) < 0.00001
    # Since `maxInc=1000`, then the following shouldn't be accepted,
    # so the value is still 0.0314
    number_lineedit.double_text = "3.14e9"
    assert (abs(number_lineedit.double_proxy._edit_binding.value - 0.0314)
            < 0.00001)
    assert number_lineedit.double_proxy.edit_value is None
    # Integers
    number_lineedit.int_text = "3"
    assert number_lineedit.int_proxy.edit_value == 3
    assert number_lineedit.int_text == "3"

    # Since 12 is greater than `maxInc=10`, then it shouldn't be accepted,
    # so the value is still 3
    number_lineedit.int_text = "12"
    assert number_lineedit.int_proxy.edit_value != "12"
    assert number_lineedit.int_proxy.edit_value is None
    assert number_lineedit.int_text == "12"


def test_number_lineedit_scientific_notation(number_lineedit_setup):
    number_lineedit = number_lineedit_setup
    test_strings = ["3.141592e0", "1.23e2", "1.23e+2", "1.23e-1"]
    results = [3.141592, 123., 123., 0.123]
    for i, test_string in enumerate(test_strings):
        number_lineedit.double_text = test_string
        assert number_lineedit.double_proxy.edit_value == results[i]
        number_lineedit.float_controller.internal_widget.setText(test_string)
        assert number_lineedit.float_proxy.edit_value == results[i]

    last_accepted_value = "1.2"
    number_lineedit.double_text = last_accepted_value
    test_out_of_range = ["1.0e5", "-1.0e5"]
    results = [100000, -100000]
    for i, test_string in enumerate(test_out_of_range):
        # shouldn't be accepted, the edit value is None:
        number_lineedit.double_text = test_string
        assert number_lineedit.double_proxy.edit_value != results[i]
        assert number_lineedit.double_proxy.edit_value is None


def test_number_lineedit_change_decimals(number_lineedit_setup, mocker):
    number_lineedit = number_lineedit_setup
    action = number_lineedit.double_controller.widget.actions()[0]
    assert "decimals" in action.text().lower()

    sym = "karabogui.controllers.edit.lineedit.QInputDialog"
    QInputDialog = mocker.patch(sym)
    QInputDialog.getInt.return_value = 4, True
    action.trigger()

    assert number_lineedit.double_controller.model.decimals == 4


def test_number_lineedit_decimal_validation(number_lineedit_setup):
    number_lineedit = number_lineedit_setup
    number_lineedit.double_text = "1.0"
    assert number_lineedit.double_proxy.edit_value == 1.0
    number_lineedit.double_controller.model.decimals = 3
    # invalid input for floating decimals
    number_lineedit.double_text = "1.0003"
    assert number_lineedit.double_proxy.edit_value is None
    number_lineedit.double_text = "1.231"
    assert number_lineedit.double_proxy.edit_value == 1.231
    number_lineedit.double_text = "1e-1"
    assert number_lineedit.double_proxy.edit_value == 0.1
    # try to trick decimals fails!
    number_lineedit.double_text = "1.278e-2"
    assert number_lineedit.double_proxy.edit_value is None
    # follow user input, we are only allowed to set 3 digits
    number_lineedit.double_text = "1"
    assert number_lineedit.double_proxy.edit_value == 1.0
    number_lineedit.double_text = "1.1"
    assert number_lineedit.double_proxy.edit_value == 1.1
    number_lineedit.double_text = "1.12"
    assert number_lineedit.double_proxy.edit_value == 1.12
    number_lineedit.double_text = "1.124"
    assert number_lineedit.double_proxy.edit_value == 1.124
    number_lineedit.double_text = "1.1244"
    assert number_lineedit.double_proxy.edit_value is None


def test_number_lineedit_decline_color(number_lineedit_setup):
    number_lineedit = number_lineedit_setup
    number_lineedit.double_text = "10000.0"
    assert number_lineedit.double_proxy.edit_value is None
    error_sheet = number_lineedit.double_controller._style_sheet.format("red")
    assert (number_lineedit.double_controller.internal_widget.styleSheet()
            == error_sheet)
    number_lineedit.double_controller.on_decline()
    normal_sheet = number_lineedit.double_controller._style_sheet.format(
        "black")
    assert (number_lineedit.double_controller.internal_widget.styleSheet()
            == normal_sheet)


class IntObject24(Configurable):
    prop = Int32(minExc=-2 ** 24,
                 maxExc=2 ** 24)  # 12345678 < 2*24 < 123456789


class NumberIntEditSetUp:
    def __init__(self):
        self.proxy = get_class_property_proxy(
            IntObject24.getClassSchema(), "prop")
        self.controller = IntLineEdit(proxy=self.proxy,
                                      model=IntLineEditModel())
        self.controller.create(None)
        self.controller.set_read_only(False)
        self.normal_sheet = self.controller._style_sheet.format("black")
        self.error_sheet = self.controller._style_sheet.format("red")

    @property
    def sheet(self):
        return self.controller.internal_widget.styleSheet()

    @property
    def text(self):
        return self.controller.internal_widget.text()

    @text.setter
    def text(self, value):
        return self.controller.internal_widget.setText(value)


def test_number_int_property_proxy_edit_values_from_text_input(gui_app):
    # setup
    number_int = test_object = NumberIntEditSetUp()
    # test body
    set_proxy_value(number_int.proxy, "prop", 1234)
    number_int.text = "12345"
    number_int.text = "123456"
    number_int.text = "1234567"
    number_int.text = "12345678"
    assert number_int.sheet == number_int.normal_sheet
    number_int.text = "123456789"
    assert number_int.sheet == number_int.error_sheet
    assert get_editor_value(number_int.proxy) != 12345678
    assert get_editor_value(number_int.proxy) == 1234
    # teardown
    test_object.controller.destroy()
    assert test_object.controller.widget is None


class HexadecimalSetUp:
    def __init__(self):
        self.proxy = get_class_property_proxy(Uint8Object.getClassSchema(),
                                              "prop")
        self.controller = Hexadecimal(proxy=self.proxy)
        self.controller.create(None)
        self.controller.set_read_only(False)
        self.normal_sheet = self.controller._style_sheet.format("black")
        self.error_sheet = self.controller._style_sheet.format("red")

    @property
    def sheet(self):
        return self.controller.internal_widget.styleSheet()

    @property
    def text(self):
        return self.controller.internal_widget.text()

    @text.setter
    def text(self, value):
        return self.controller.internal_widget.setText(value)


def test_hexadecimal_set_value(gui_app):
    # setup
    hexadecimal_edit = HexadecimalSetUp()
    # test body
    set_proxy_value(hexadecimal_edit.proxy, "prop", 0x40)
    hexadecimal_edit.text = ""
    assert hexadecimal_edit.sheet == hexadecimal_edit.error_sheet

    hexadecimal_edit.text = "40"
    assert hexadecimal_edit.text == "40"
    assert hexadecimal_edit.sheet == hexadecimal_edit.normal_sheet

    hexadecimal_edit.text = "-40"
    assert hexadecimal_edit.text == "-40"
    assert get_editor_value(hexadecimal_edit.proxy) != -0x40
    assert hexadecimal_edit.sheet == hexadecimal_edit.error_sheet

    hexadecimal_edit.text = "8F"
    assert hexadecimal_edit.sheet == hexadecimal_edit.normal_sheet
    assert get_editor_value(hexadecimal_edit.proxy) == 0x8F
    # teardown
    hexadecimal_edit.controller.destroy()
    assert hexadecimal_edit.controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QKeyEvent

from karabogui.dialogs.api import AboutDialog
from karabogui.testing import GuiTestCase, singletons


class TestAboutDialog(GuiTestCase):

    def keyKeyPressEvent(self, dialog, key):
        event = QKeyEvent(QEvent.KeyPress, 0, Qt.NoModifier,
                          key, False, 1)
        dialog.keyPressEvent(event)

    def test_basic_dialog(self):
        network = mock.Mock()
        manager = mock.Mock()
        with singletons(network=network, manager=manager):
            dialog = AboutDialog()
            for key in "chooch":
                self.keyKeyPressEvent(dialog, key)
            network.togglePerformanceMonitor.assert_called_once()

            network.reset_mock()
            for key in "cshooch":
                self.keyKeyPressEvent(dialog, key)
            network.togglePerformanceMonitor.assert_not_called()

            network.reset_mock()
            for key in "achooch":
                self.keyKeyPressEvent(dialog, key)
            network.togglePerformanceMonitor.assert_called_once()

            for key in "bigdata":
                self.keyKeyPressEvent(dialog, key)
            manager.toggleBigDataPerformanceMonitor.assert_called_once()

            manager.reset_mock()
            for key in "bigndata":
                self.keyKeyPressEvent(dialog, key)
            manager.toggleBigDataPerformanceMonitor.assert_not_called()


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QRect
from traits.api import HasStrictTraits, List

from .tools.api import is_resizable, is_selectable
from .utils import calc_bounding_rect


class SceneSelectionModel(HasStrictTraits):
    """A selection model for the SceneView.
    """
    # The list of selected scene objects (shapes, widgets, layouts)
    _selection = List

    def __iter__(self):
        """Implement the Python iterator interface.
        """
        return iter(self._selection)

    def __len__(self):
        return len(self._selection)

    def clear_selection(self):
        """Remove all objects from the selection
        """
        self._selection = []

    def deselect_object(self, obj):
        """Deselect an object.
        """
        self._selection.remove(obj)

    def get_selection_bounds(self):
        """Return the bounding rectangle for the objects which are selected.
        """
        return self.get_item_rect()

    def get_item_rect(self):
        """Return the item rectangle for the objects which are selected.

        This rect is used for position determination and not for drawing!
        """
        return QRect(*calc_bounding_rect(self._selection))

    def has_selection(self):
        return len(self._selection) > 0

    def has_multiple_selection(self):
        return len(self._selection) > 1

    def select_object(self, obj):
        """Select an object that can be handled by the scene tools"""
        if is_selectable(obj):
            self._selection.append(obj)

    def child_in_rect(self, rect):
        for child in self._selection:
            if child.geometry().intersects(rect):
                return child

    def child_in_pos(self, pos):
        for child in self._selection:
            if child.geometry().contains(pos):
                return child

    def get_single_selection(self):
        if len(self._selection) == 1:
            return self._selection[0]

    def has_resizable_selection(self):
        """Evaluates if the selection is resizable. The rule is:
           1. There is only one selection object in the scene
           2. That object can be resized."""
        gui_obj = self.get_single_selection()
        return gui_obj is not None and is_resizable(gui_obj)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
"""Custom API definition for Qsci autocompletion"""

global _KARABO_NAMESPACE
_KARABO_NAMESPACE = None


def get_symbols():
    global _KARABO_NAMESPACE
    if _KARABO_NAMESPACE is None:
        _KARABO_NAMESPACE = create_symbols()
    return _KARABO_NAMESPACE


_MIDDLELAYER = (
    "allCompleted",
    "background",
    "call",
    "callNoWait",
    "connectDevice",
    "execute",
    "executeNoWait",
    "DeviceClientBase",
    "firstCompleted",
    "getDevice",
    "getDevices",
    "getConfiguration",
    "getInstanceInfo",
    "getSchema",
    "getServers",
    "getTimeInfo",
    "getTopology",
    "get_property",
    "instantiate",
    "isAlive",
    "Macro",
    "MacroSlot",
    "PipelineContext",
    "set_property",
    "setWait",
    "setNoWait",
    "shutdown",
    "sleep",
    "updateDevice",
    "waitUntil",
    "waitUntilNew",
    "waitWhile",
)

_ASYNCIO = (
    "CancelledError",
    "gather",
    "sleep",
    "TimeoutError",
    "wait_for",
)


def create_symbols():
    """Lazily create all symbols for the global namespace of a Scintilla
    editor
    """
    import karabo.native as native
    native_symbols = [f"karabo.middlelayer.{symbol}"
                      for symbol in dir(native) if not symbol.startswith("__")]

    mdl_symbols = [f"karabo.middlelayer.{symbol}" for symbol in _MIDDLELAYER]

    asyncio_symbols = [f"asyncio.{symbol}" for symbol in _ASYNCIO]
    symbols = native_symbols + mdl_symbols + asyncio_symbols

    return symbols



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 10, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QDialog, QPushButton
from traits.api import Instance

from karabo.common.scenemodel.api import EditableListElementModel
from karabogui.binding.api import VectorStringBinding, get_editor_value
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)
from karabogui.dialogs.listedit import ListEditDialog


@register_binding_controller(ui_name="List Element Field", can_edit=True,
                             klassname="EditableListElement",
                             binding_type=VectorStringBinding)
class EditableListElement(BaseBindingController):
    model = Instance(EditableListElementModel, args=())

    def create_widget(self, parent):
        widget = QPushButton("Edit list", parent)
        widget.setStyleSheet("QPushButton { text-align: center; }")
        widget.clicked.connect(self._on_edit_clicked)
        return widget

    def _on_edit_clicked(self):
        binding = self.proxy.binding
        if binding is None:
            return

        list_edit = ListEditDialog(binding, duplicates_ok=True,
                                   parent=self.widget)
        list_edit.set_list(get_editor_value(self.proxy, []))
        if list_edit.exec() == QDialog.Accepted:
            self.proxy.edit_value = list_edit.values

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.singletons.mediator import Mediator
from karabogui.testing import singletons
from karabogui.util import process_qt_events


def test_basic_mediator(gui_app):
    instanceId = ""

    def _show_device(data):
        nonlocal instanceId
        instanceId = data.get("instanceId", "")

    event_map = {
        KaraboEvent.ShowDevice: _show_device,
    }

    mediator = Mediator()
    assert len(mediator._listeners.keys()) == 0

    with singletons(mediator=mediator):
        register_for_broadcasts(event_map)
        assert KaraboEvent.ShowDevice in mediator._listeners.keys()
        assert len(mediator._listeners.keys()) == 1

        assert instanceId == ""
        broadcast_event(KaraboEvent.ShowDevice,
                        data={"instanceId": "Marty"})
        process_qt_events(timeout=10)
        assert instanceId == "Marty"

        # Unregister once
        unregister_from_broadcasts(event_map)
        assert len(mediator._listeners.keys()) == 0

        # Unregister twice, no harm
        unregister_from_broadcasts(event_map)
        assert len(mediator._listeners.keys()) == 0


if __name__ == "__main__":
    pytest.main()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 6, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QRect, QSize
from qtpy.QtWidgets import QBoxLayout, QGridLayout, QLayout, QWidgetItem

from karabogui.sceneview.utils import calc_bounding_rect

from .base import BaseLayout


class GroupLayout(BaseLayout, QLayout):
    def __init__(self, model, parent=None):
        super().__init__(model, parent)
        self._children = []
        self._child_bounds = None

    def _add_layout(self, layout):
        self.addItem(layout)
        super()._add_layout(layout)

    def _add_widget(self, widget):
        self.addWidget(widget)

    # --------------------------------------------
    # QLayout Virtual Functions

    def addItem(self, item):
        self._children.append(item)
        self._child_bounds = None

    def removeItem(self, item):
        self._children.remove(item)
        self._child_bounds = None
        super().removeItem(item)

    def itemAt(self, index):
        try:
            return self._children[index]
        except IndexError:
            return

    def takeAt(self, index):
        item = self._children.pop(index)
        self._child_bounds = None
        layout = item.layout()
        if layout is not None and layout.parent() is self:
            layout.setParent(None)
        # The creation path of the layout items bypasses the virtual
        # wrapper methods, this means that the ownership of the cpp
        # pointer is never transfered to Qt. If the item is returned
        # here it will be delete by Qt, which doesn't own the pointer.
        # A double free occurs once the Python item falls out of scope.
        # To avoid this, this method always returns None and the item
        # cleanup is performed by Python, which owns the cpp pointer.
        del item
        return None

    def count(self):
        return len(self._children)

    def geometry(self):
        model = self.model
        rect = QRect(model.x, model.y, model.width, model.height)
        if rect.isEmpty():
            rect = QRect(*self._child_bounding_rect())
        return rect

    def maximumSize(self):
        return self.sizeHint()

    def minimumSize(self):
        return self.sizeHint()

    def sizeHint(self):
        x, y, w, h = self._child_bounding_rect()
        return QSize(w, h)

    def setGeometry(self, rect):
        curr_geometry = self.geometry()
        if rect == curr_geometry:
            # NOTE: If we don't bail out here, Qt will continue to call us!
            # I have no idea why that is, but it will eat your CPU!
            return

        # Translate all children by the amount our origin moved
        offset = rect.topLeft() - curr_geometry.topLeft()
        for item in self:
            if isinstance(item, QWidgetItem):
                item.widget().translate(offset)
            else:
                item.translate(offset)
        # Invalidate the cached bounds!
        self._child_bounds = None

        super().setGeometry(rect)

    def _child_bounding_rect(self):
        """Cache the result of calc_bounding_rect to avoid expensive
        recalculation.
        """
        if self._child_bounds is None:
            self._child_bounds = calc_bounding_rect(self._children)
        return self._child_bounds


VERTICAL_LAYOUTS = (QBoxLayout.TopToBottom, QBoxLayout.BottomToTop)


class BoxLayout(BaseLayout, QBoxLayout):
    def __init__(self, model, direction, parent=None):
        super().__init__(model, direction, parent)
        self.setContentsMargins(0, 0, 0, 0)

    def _add_layout(self, layout):
        self.addItem(layout)
        super()._add_layout(layout)

    def _add_widget(self, widget):
        self.addWidget(widget)

    def minimumSize(self):
        # We don't set a minimum size on the layout to enable more flexible
        # resizing
        return QSize()

    def sizeHint(self):
        """Calculate the sizeHint of the layout from the sizeHints of the
           children. This varies with direction."""

        # We compute the sizeHint only if the model is invalid.
        # We use the child bounding rects if otherwise.
        model = self.model
        if QRect(model.x, model.y, model.width, model.height).isEmpty():
            sizes = [self.itemAt(i).sizeHint() for i in range(self.count())]
            widths = [size.width() for size in sizes]
            heights = [size.height() for size in sizes]

            get_width, get_height = sum, max
            if self.direction() in VERTICAL_LAYOUTS:
                get_width, get_height = get_height, get_width

            return QSize(get_width(widths), get_height(heights))

        return super().sizeHint()

    def setGeometry(self, rect):
        super().setGeometry(rect)

        # Update the geometry of the widget in their model
        for i in range(self.count()):
            item = self.itemAt(i)
            rect = item.geometry()

            if isinstance(item, BaseLayout):
                model = item.model
            elif isinstance(item, QWidgetItem):  # QWidgetItem
                model = item.widget().model
            else:
                # This is probably a ShapeLayoutItem. It has a `set_geometry`
                # that sets the values to its model. We do nothing instead.
                return

            # Update the model geometry
            model.trait_set(x=rect.x(), y=rect.y(),
                            width=rect.width(), height=rect.height())


class GridLayout(BaseLayout, QGridLayout):
    def __init__(self, model, parent=None):
        super().__init__(model, parent)

    def _add_layout(self, layout):
        ld = layout.model.layout_data
        self.addLayout(layout, ld.row, ld.col, ld.rowspan, ld.colspan)
        super()._add_layout(layout)

    def _add_widget(self, widget):
        ld = widget.model.layout_data
        self.addWidget(widget, ld.row, ld.col, ld.rowspan, ld.colspan)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QDialog, QFrame, QLabel
from traits.api import Instance, String, Tuple, Undefined

from karabo.common.scenemodel.api import build_model_config
from karabogui.binding.api import (
    FloatBinding, get_binding_format, get_binding_value)
from karabogui.controllers.base import BaseBindingController
from karabogui.controllers.unitlabel import add_unit_label
from karabogui.dialogs.api import (
    AlarmDialog, FormatFmtDialog, FormatLabelDialog)
from karabogui.fonts import get_font_size_from_dpi
from karabogui.indicators import (
    ALL_OK_COLOR, PROPERTY_ALARM_COLOR, PROPERTY_WARN_COLOR)
from karabogui.util import generateObjectName
from karabogui.widgets.hints import Label


class BaseLabelController(BaseBindingController):
    model = Undefined

    bg_color = Tuple(ALL_OK_COLOR)
    internal_widget = Instance(QLabel)
    style_sheet = String
    fmt = String("{}")

    def create_widget(self, parent):
        self.internal_widget = Label(parent)
        widget = add_unit_label(self.proxy, self.internal_widget,
                                parent=parent)
        widget.setFrameStyle(QFrame.Box | QFrame.Plain)

        objectName = generateObjectName(self)
        self.style_sheet = (f"QWidget#{objectName}" +
                            " {{ background-color : rgba{}; }}")
        widget.setObjectName(objectName)
        sheet = self.style_sheet.format(ALL_OK_COLOR)
        widget.setStyleSheet(sheet)

        # Add an action for formatting options
        format_action = QAction("Format field", widget)
        format_action.triggered.connect(self._format_field)
        widget.addAction(format_action)
        self._apply_format(widget)

        return widget

    def clear_widget(self):
        """Clear the internal widget when the device goes offline"""
        self.internal_widget.clear()

    def binding_update(self, proxy):
        self.binding_update_proxy(proxy)
        self.widget.update_unit_label(proxy)

    def value_update(self, proxy):
        binding = proxy.binding
        value = get_binding_value(proxy, "")
        if isinstance(value, str):
            # Early bail out for Long binary data (e.g. image) or
            # if the property is not set (Undefined)
            self.internal_widget.setText(value[:255])
            return

        self.value_update_proxy(proxy, value)
        # Provide some pretty value formatting
        if isinstance(binding, FloatBinding):
            value = float(str(value))

        self.internal_widget.setText(self.toString(value))

    # -----------------------------------------------------------------------
    # Subclassing methods

    def binding_update_proxy(self, proxy):
        """Subclass this method to react on a binding update of `proxy`"""

    def value_update_proxy(self, proxy, value):
        """Subclass this method to react on a value update of `proxy`"""

    def toString(self, value):
        """Subclass this method to convert on a value to string"""
        try:
            value = self.fmt.format(value)
        except Exception:
            value = str(value)
        return value

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        """The widget is passed as an argument in create_widget as it is not
           yet bound to self.widget then"""
        if widget is None:
            widget = self.widget

        # Apply font formatting
        font = widget.font()
        font.setPointSize(get_font_size_from_dpi(self.model.font_size))
        font.setBold(self.model.font_weight == "bold")
        widget.setFont(font)


# --------------------------------------------------------------------


class FormatMixin:
    """The FormatMixin to provide string formatting for floating points"""
    def create_widget(self, parent):
        widget = super().create_widget(parent)

        fmt_action = QAction("Format value", widget)
        fmt_action.triggered.connect(self._fmt_expression)
        widget.addAction(fmt_action)
        self._create_fmt()

        return widget

    @classmethod
    def initialize_model(cls, proxy, model):
        """Initialize the formatting from the binding of the proxy"""
        super().initialize_model(proxy, model)
        fmt, decimals = get_binding_format(proxy.binding)
        model.trait_set(fmt=fmt, decimals=decimals)

    # -----------------------------------------------------------------------
    # Formatting methods

    def _fmt_expression(self):
        value = get_binding_value(self.proxy)
        dialog = FormatFmtDialog(
            value, fmt=self.model.fmt, decimals=self.model.decimals,
            parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(
                fmt=dialog.fmt, decimals=dialog.decimals)
            self._create_fmt()
            self.value_update(self.proxy)

    def _create_fmt(self):
        """Generate the string formatting according to the model settings"""
        self.fmt = f"{{:.{self.model.decimals}{self.model.fmt}}}"


class AlarmMixin:
    """The AlarmMixin to provide alarm configuration for numbers"""

    def create_widget(self, parent):
        widget = super().create_widget(parent)
        alarm_action = QAction("Configure alarms", widget)
        alarm_action.triggered.connect(self._alarm_dialog)
        widget.addAction(alarm_action)

        return widget

    def value_update_proxy(self, proxy, value):
        model = self.model
        alarm_low = model.alarmLow
        alarm_high = model.alarmHigh
        warn_low = model.warnLow
        warn_high = model.warnHigh
        if ((alarm_low is not Undefined and value < alarm_low) or
                (alarm_high is not Undefined and value > alarm_high)):
            self.bg_color = PROPERTY_ALARM_COLOR
        elif ((warn_low is not Undefined and value < warn_low) or
              (warn_high is not Undefined and value > warn_high)):
            self.bg_color = PROPERTY_WARN_COLOR
        else:
            self.bg_color = ALL_OK_COLOR
        sheet = self.style_sheet.format(self.bg_color)
        self.widget.setStyleSheet(sheet)

    def _alarm_dialog(self):
        binding = self.proxy.binding
        config = build_model_config(self.model)
        dialog = AlarmDialog(binding, config, parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(**dialog.values)
            self.value_update(self.proxy)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import PlotItem
from qtpy.QtWidgets import QMenu
from traits.api import (
    Bool, Dict, Event, HasStrictTraits, Instance, Property, String, Type)

from .analyzer import BaseAnalyzer
from .plot import BasePlot


class ControllerAggregator(HasStrictTraits):
    """Aggregates multiple plot controllers to a unified interface for the base
       image plot to communicate with."""

    # Override to specify the plot class and its orientation
    # `{orientation: plot_klass}`
    contents = Dict

    # The dictionary which contains the instantiated aux plot controllers
    # `{orientation: plot_klass()}`
    controllers = Dict

    # List of plotItems from the aux plots
    plotItems = Property

    # Show the stats on the labelItem
    show_stats = Bool(False)

    # Forwards the stats to the aux plot controller
    stats = Event

    # Shared menu for the aux plot contents
    menu = Property(Instance(QMenu))

    def __init__(self, config=None, **traits):
        super().__init__(**traits)

        for orientation, plot_klass in self.contents.items():
            if config is None:
                config = {}
            controller = plot_klass(orientation=orientation, **config)
            self.controllers[orientation] = controller

    # -----------------------------------------------------------------------
    # Public methods

    def process(self, region):
        # The default process implementation. No fancy config needed
        for plot in self.controllers.values():
            plot.process(region)

    def link_view(self, view):
        """Connects the image plot viewbox to the respective aux plots"""

    def set_axes(self, x_data, y_data):
        """Set the transformed image axes data to the respective aux plots"""

    def destroy(self):
        """Destroy the children members and disconnect trait handlers"""
        for controller in self.controllers.values():
            controller.destroy()

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_plotItems(self):
        return [plot.plotItem for plot in self.controllers.values()]

    def _get_menu(self):
        """See if there's shared menu"""


class BaseController(HasStrictTraits):
    """Aux plot controller."""

    # The analyzer class performs the calculations/data processing.
    # This is usually overwritten and initialized on the subclass
    analyzer = Instance(BaseAnalyzer)

    # The plot class is overwritten on the subclass to specify which plot
    # item controller to use.
    plot_klass = Type(BasePlot)

    # The plot item controller class. This is instantiated with the specified
    # `plot_klass` from the subclass.
    plot = Instance(BasePlot)

    # The pyqtgraph plotItem from the `plotItem`.
    plotItem = Property(Instance(PlotItem))

    # The orientation of the plot. ["top", "left", "right", "bottom"]
    orientation = String

    def __init__(self, **traits):
        super().__init__(**traits)
        self.plot = self.plot_klass(orientation=self.orientation)

    # -----------------------------------------------------------------------
    # Public methods

    def process(self, region, **config):
        """Process is made up of two parts:
           1. Analyze the input data
           2. Plot the resulting data"""

        if not region.valid():
            self.clear_data()
            return

        x_data, y_data = self.analyzer.analyze(region)
        self.plot.set_data(x_data, y_data)

    def clear_data(self):
        self.plot.clear_data()
        self.analyzer.clear_data()

    def link_view(self, axis=0, view=None):
        """Link the image plot viewbox and the aux plot wrt specified axis"""
        self.plotItem.vb.linkView(axis, view)

    def destroy(self):
        """Destroy the children cleanly"""
        self.plot.destroy()

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_plotItem(self):
        return self.plot.plotItem



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import QStringListModel, Slot
from qtpy.QtWidgets import QDialog, QMessageBox

from karabogui import icons, messagebox

from .utils import get_dialog_ui


class ProxiesDialog(QDialog):
    """The basic dialog to remove additional proxies from a `container`
    """

    def __init__(self, container, parent=None):
        super().__init__(parent)
        self.setModal(False)
        ui_file = get_dialog_ui("proxies_dialog.ui")
        uic.loadUi(ui_file, self)

        self.widget_container = container
        controller = container.widget_controller
        self.ui_main_property.setText(controller.proxy.key)

        list_model = QStringListModel()
        paths = [proxy.key for proxy in controller.proxies[1:]]
        list_model.setStringList(paths)
        self._list_model = list_model

        self.ui_list_view.setModel(list_model)
        # Removal button
        self.ui_remove_button.clicked.connect(self._on_remove_clicked)
        self.ui_remove_button.setIcon(icons.delete)
        self._on_update_buttons()

    # ----------------------------------------------------------------------
    # Private interface

    def _on_update_buttons(self):
        """Keep the removal button in sync with the items"""
        has_items = self._list_model.rowCount() > 0
        self.ui_remove_button.setEnabled(has_items)

    # ----------------------------------------------------------------------
    # Slots

    @Slot()
    def _on_remove_clicked(self):
        index = self.ui_list_view.selectionModel().currentIndex()
        if not index.isValid():
            return

        controller = self.widget_container.widget_controller
        proxy = controller.proxies[index.row() + 1]
        prop = proxy.key
        ask = f"Are you sure you want to remove the property <b>{prop}</b>?"
        options = (QMessageBox.Yes | QMessageBox.No)
        reply = QMessageBox.question(self, "Remove Property", ask, options,
                                     QMessageBox.No)
        if reply == QMessageBox.No:
            return

        if not self.widget_container.remove_proxies([proxy]):
            messagebox.show_error(f"The removal of property {prop} is not "
                                  "supported by this controller.",
                                  parent=self.parent())
            return
        self._list_model.setStringList([proxy.key for proxy
                                        in controller.proxies[1:]])
        self._on_update_buttons()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtWidgets import QDialog

from karabogui.dialogs.utils import get_dialog_ui
from karabogui.singletons.api import get_config


class DebugRunDialog(QDialog):
    """
    A dialog to show the list of MacroServers with 'Development' serverFlag,
    in a combobox.
    """

    def __init__(self, serverIds=None, parent=None):
        super().__init__(parent=parent)
        ui_file = get_dialog_ui("debug_run.ui")
        uic.loadUi(ui_file, self)
        button_box = self.buttonBox
        run_button = button_box.button(button_box.Ok)
        run_button.setText("Run")
        if not serverIds:
            run_button.setEnabled(False)
            run_button.setToolTip(
                "Make sure to have at least one Development MacroServer")
        else:
            self.comboBox.addItems(serverIds)

            previous_server = get_config()["macro_development"]
            if previous_server in serverIds:
                self.comboBox.setCurrentText(previous_server)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on November 13, 2024
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QPalette
from qtpy.QtWidgets import (
    QDialog, QHeaderView, QListWidgetItem, QTableWidget, QTableWidgetItem,
    QWidget)

from karabo.native import Hash, Timestamp
from karabogui import icons
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.request import call_device_slot
from karabogui.topology.api import getTopology
from karabogui.util import get_spin_widget

from .utils import get_dialog_ui

BLANK_PAGE = 0
WAITING_PAGE = 1
INFO_PAGE = 2


class GuiSessionInfo(QDialog):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        uic.loadUi(get_dialog_ui("guiserverinfo.ui"), self)

        self.set_stack_index(BLANK_PAGE)

        # Stacked widget parameters
        self.stacked_widget.addWidget(QTableWidget(self))

        wait_widget = get_spin_widget(icon="wait", parent=self)
        wait_widget.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        wait_widget.setAutoFillBackground(True)
        wait_widget.setBackgroundRole(QPalette.Base)
        self.stacked_widget.addWidget(wait_widget)

        self.table_widget = QTableWidget(self)
        self.stacked_widget.addWidget(self.table_widget)

        self.fill_servers()
        self.list_widget.itemClicked.connect(self.itemClicked)
        self.table_widget.setColumnCount(3)
        self.table_widget.setHorizontalHeaderLabels(
            ["Version", "Session start", "Session token"])

        header = self.table_widget.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.Stretch)

        self.refresh_button.setIcon(icons.refresh)
        self.refresh_button.clicked.connect(self.refresh)

        self._event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network}
        register_for_broadcasts(self._event_map)

    def fill_servers(self):
        topology = getTopology()
        devices = []
        for deviceId, _, attrs in topology.get("device", Hash()).iterall():
            if attrs.get("classId") == "GuiServerDevice":
                devices.append(deviceId)
        if devices:
            self.list_widget.addItems(devices)
            self.list_widget.setCurrentRow(0)
            self.itemClicked(self.list_widget.currentItem())

    def set_stack_index(self, index):
        self.stacked_widget.blockSignals(True)
        self.stacked_widget.setCurrentIndex(index)
        self.stacked_widget.blockSignals(False)

    @Slot()
    def refresh(self):
        item = self.list_widget.currentItem()
        if item is not None:
            self.itemClicked(item)

    @Slot(QListWidgetItem)
    def itemClicked(self, item):

        def handler(success, reply):
            if not success:
                self.set_stack_index(BLANK_PAGE)
                return

            self.set_stack_index(INFO_PAGE)
            sessions = reply["clientSessions"]
            self.table_widget.setRowCount(len(sessions))
            for row, session in enumerate(sessions):
                item = QTableWidgetItem(session["clientVersion"])
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                self.table_widget.setItem(row, 0, item)

                timestamp = Timestamp(
                    session["sessionStartTime"]).toLocal(" ")
                item = QTableWidgetItem(timestamp)
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                self.table_widget.setItem(row, 1, item)

                item = QTableWidgetItem(session["sessionToken"])
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                self.table_widget.setItem(row, 2, item)

        self.set_stack_index(WAITING_PAGE)
        deviceId = item.data(Qt.DisplayRole)
        call_device_slot(handler, deviceId, "slotGetClientSessions")

    def done(self, result):
        unregister_from_broadcasts(self._event_map)
        return super().done(result)

    def _event_network(self, data):
        if not data.get("status"):
            self.close()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 23, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import webbrowser

from qtpy.QtCore import QPoint, QRect, QRectF, QSize, Qt, Slot
from qtpy.QtGui import QColor, QPainter, QPen
from qtpy.QtWidgets import (
    QAction, QDialog, QLabel, QPushButton, QStackedLayout)

from karabo.common.scenemodel.api import SceneTargetWindow
from karabogui import messagebox
from karabogui.binding.api import ProxyStatus
from karabogui.dialogs.api import (
    DeviceCapabilityDialog, SceneLinkDialog, TextDialog, WebDialog)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.fonts import get_qfont
from karabogui.indicators import get_device_status_pixmap
from karabogui.request import get_scene_from_server
from karabogui.singletons.api import get_topology
from karabogui.topology.api import is_device_online
from karabogui.widgets.hints import KaraboSceneWidget


class BaseLinkWidget(KaraboSceneWidget, QPushButton):
    """The base link widget for the scene"""

    iconColor = None

    def __init__(self, model, parent=None):
        super().__init__(model=model, parent=parent)
        self.setFont(get_qfont(model.font))
        self.setToolTip(self.getToolTip())
        self.setCursor(Qt.PointingHandCursor)
        self.clicked.connect(self.handle_click)
        self.setGeometry(QRect(model.x, model.y, model.width, model.height))

        edit_action = QAction("Edit Label", self)
        edit_action.triggered.connect(self.edit_label)
        self.addAction(edit_action)

        target_action = QAction("Edit Target", self)
        target_action.triggered.connect(self.edit_target)
        self.addAction(target_action)

    def set_geometry(self, rect):
        """Satisfy the informal widget interface."""
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        """Satisfy the informal widget interface."""
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def paintEvent(self, event):
        with QPainter(self) as painter:
            boundary = self.rect().adjusted(0, 0, -1, -1)
            pt = boundary.topLeft()
            rects = [QRectF(QRect((pt), QSize(7, 7))),
                     QRectF(QRect((pt) + QPoint(11, 0), QSize(7, 7)))]
            painter.fillRect(boundary, QColor(self.model.background))
            pen = QPen(Qt.black)
            pen.setWidth(self.model.frame_width)
            painter.setPen(pen)
            painter.drawRect(boundary)
            pen.setColor(self.iconColor)
            pen.setWidth(3)
            painter.setPen(pen)
            painter.drawRects(rects)
            pen.setColor(Qt.lightGray)
            painter.setPen(pen)
            painter.drawLine(pt + QPoint(4, 4), pt + QPoint(15, 4))
            # Before painting the text, set the font
            painter.setFont(self.font())
            pen = QPen(QColor(self.model.foreground))
            painter.setPen(pen)
            painter.drawText(boundary, Qt.AlignCenter, self.model.text)

    def edit(self, scene_view=None):
        """Double click handle for the scene view"""
        self.on_edit()
        self.setToolTip(self.getToolTip())

    # Qt Slots
    # -----------------------------------------------------------------------

    @Slot()
    def edit_label(self):
        dialog = TextDialog(self.model, parent=self)
        if dialog.exec() == QDialog.Rejected:
            return

        # Set all at once!
        label = dialog.label_model
        self.model.trait_set(text=label.text, frame_width=label.frame_width,
                             font=label.font, background=label.background,
                             foreground=label.foreground)

        # Record the QFont on the widget
        self.setFont(get_qfont(label.font))
        self.update()

    @Slot()
    def edit_target(self):
        self.on_edit()

    @Slot()
    def handle_click(self):
        self.on_click()

    # Public abstract interface
    # -----------------------------------------------------------------------

    def getToolTip(self):
        return self.model.target

    def on_edit(self):
        """Subclass this method to provide an action on `edit`"""

    def on_click(self):
        """Subclass this method to provide an action on `click`"""


class SceneLinkWidget(BaseLinkWidget):
    """A clickable link which opens another `scene`"""
    iconColor = Qt.darkGray

    def on_click(self):
        """Reimplemented function of `BaseLinkWidget`"""
        if len(self.model.target) > 0:
            parts = self.model.target.split(":")
            if len(parts) != 2:
                return

            # target format => "simple_name:UUID"
            name = parts[0]
            target = parts[1]
            target_window = self.model.target_window
            # Broadcast an event
            data = {"name": name, "target": target,
                    "target_window": target_window}
            broadcast_event(KaraboEvent.OpenSceneLink, data)

    def on_edit(self):
        """Reimplemented function of `BaseLinkWidget"""
        dialog = SceneLinkDialog(self.model, parent=self)
        if dialog.exec() == QDialog.Rejected:
            return

        self.model.target = dialog.selectedScene
        self.model.target_window = dialog.selectedTargetWindow


class WebLinkWidget(BaseLinkWidget):
    """A clickable widget which opens a hyperlink
    """
    iconColor = QColor(255, 145, 255)

    def on_click(self):
        """Reimplemented function of `BaseLinkWidget`"""
        if len(self.model.target) > 0:
            try:
                webbrowser.open_new(self.model.target)
            except webbrowser.Error:
                messagebox.show_error("No web browser available!", parent=self)

    def on_edit(self):
        """Reimplemented function of `BaseLinkWidget`"""
        dialog = WebDialog(self.model.target, parent=self)
        if dialog.exec() == QDialog.Rejected:
            return
        self.model.target = dialog.target


def _get_device_id(keys):
    if not isinstance(keys, list):
        return ''
    try:
        return keys[0].split('.', 1)[0]
    except IndexError:
        return ''


class DeviceSceneLinkWidget(BaseLinkWidget):
    """A clickable widget which opens a Device Scene Link
    """
    iconColor = QColor(100, 149, 237)

    def __init__(self, model, parent=None):
        super().__init__(model, parent)
        self.status_symbol = QLabel("", self)
        self.status_symbol.setAttribute(Qt.WA_TransparentForMouseEvents)
        pixmap = get_device_status_pixmap(ProxyStatus.OFFLINE)
        self.status_symbol.setPixmap(pixmap)

        self.layout = QStackedLayout(self)
        self.layout.setStackingMode(QStackedLayout.StackAll)
        self.layout.addWidget(self.status_symbol)

        self.proxy = get_topology().get_device(self.deviceId, False)
        self.proxy_status_change(self.proxy.status)
        self.proxy.on_trait_change(self.proxy_status_change, "status")

    @property
    def deviceId(self):
        return _get_device_id(self.model.keys)

    def proxy_status_change(self, status):
        if status is ProxyStatus.OFFLINE:
            self.status_symbol.show()
        else:
            self.status_symbol.hide()

    def getToolTip(self):
        """Reimplemented function of `BaseLinkWidget`"""
        return f"{self.deviceId}|{self.model.target}"

    def destroy(self):
        """Reimplemented function of `BaseLinkWidget`"""
        self.proxy.on_trait_change(self.proxy_status_change, "status",
                                   remove=True)
        self.proxy = None
        super().destroy()

    def on_click(self):
        """Reimplemented function of `BaseLinkWidget`"""
        device_id = self.deviceId
        if not is_device_online(device_id):
            messagebox.show_warning(
                f"Device <b>{device_id}</b> is not online!",
                "Warning", parent=self)
            return

        scene_name = self.model.target
        target_window = self.model.target_window
        get_scene_from_server(device_id, scene_name=scene_name,
                              target_window=target_window)

    def on_edit(self):
        """Reimplemented function of `BaseLinkWidget"""
        dialog = DeviceCapabilityDialog(self.deviceId, parent=self)
        dialog.find_and_select(self.model.target)
        if dialog.exec() == QDialog.Rejected:
            return
        self.model.target = dialog.capa_name
        self.model.target_window = SceneTargetWindow.Dialog



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from ast import literal_eval

import numpy as np
from traits.api import TraitError

from karabo.common import const
from karabo.native import Hash, HashList, has_changes
from karabogui.binding.binding_types import (
    BindingRoot, BoolBinding, ByteArrayBinding, CharBinding, FloatBinding,
    IntBinding, NodeBinding, SlotBinding, StringBinding, VectorBinding,
    VectorDoubleBinding, VectorFloatBinding, VectorHashBinding,
    VectorNumberBinding, VectorStringBinding)
from karabogui.binding.util import get_numpy_binding, realign_hash

VECTOR_FLOAT_BINDINGS = (VectorFloatBinding, VectorDoubleBinding)


def sanitize_table_value(binding, value):
    """Sanitize a hash list `value` against existing vector hash `binding`

    :param binding: The existing `VectorHashBinding`
    :param value: the table value (`HashList`)

    :returns: success (bool), sanitized value (`HashList`)
    """
    if value is None:
        # Theoretically this should not happen. No `None` value is sent via
        # network.
        return [], False

    msg = "Expected a value of type `HashList`, got %s instead" % type(value)
    assert isinstance(value, (list, HashList)), msg

    # Provide a success information if we received a valid table!
    success = True

    def _sanitize_row(row_bindings, row_hash):
        """Validate a single row of the table"""
        nonlocal success

        ret = Hash()
        if list(row_bindings.keys()) != list(row_hash.keys()):
            row_hash = realign_hash(row_hash, reference=row_bindings.keys())

        for path, value in row_hash.items():
            binding = row_bindings.get(path, None)
            if binding is None:
                # Tables might lose a property in a row schema. This is fine
                # as we can simply ignore this case. We don't report this as
                # invalid but continue gracefully ...
                continue

            if value is None:
                # The property is a `None` value or is not existent, e.g.
                # a column can have been added.
                # Use the default value from the binding, if necessary force!
                value = get_default_value(binding, force=True)
                success = False
            else:
                try:
                    value = binding.validate_trait("value", value)
                    # Use the fast path validate
                except TraitError:
                    value = None
                # Note: This is of course critical, if an apple becomes an
                # orange, we should have at least an orange default value.
                if value is None:
                    value = get_default_value(binding, force=True)
                    success = False
            ret[path] = value

        return ret

    ret = HashList()
    for row_hash in value:
        ret.append(_sanitize_row(binding.bindings, row_hash))

    return success, ret


def validate_value(binding, value):
    if type(value) is str:
        # There might be cases that values are saved as strings. This typically
        # occurs for table elements.
        if not isinstance(binding, (StringBinding, CharBinding)):
            value, _ = convert_string(value)
    try:
        if isinstance(binding, VectorNumberBinding):
            # Check if binding is a vector, then test array against its
            # expected `dtype`.
            casted_value = binding.check(value)
            if isinstance(binding, VECTOR_FLOAT_BINDINGS):
                value = np.array(value, dtype=casted_value.dtype)
            if not has_changes(value, casted_value):
                value = casted_value
            else:
                value = None
        elif isinstance(binding, VectorHashBinding):
            # VectorHashBinding is not a valid value and has to be explicitly
            # handled elsewhere, e.g. `validate_binding_configuration`
            value = None
        elif isinstance(binding, NodeBinding):
            # Nothing to do here! We automatically return the value
            pass
        else:
            # The value is a simple data type. We validate it with the binding
            # traits.
            value = binding.check(value)
    except (TraitError, TypeError, ValueError):
        # TraitError happens when traits cannot cast, e.g. str to bool
        # Validation has failed, we inform that there's no validated value
        value = None

    return value


def validate_table_value(binding, value):
    """Validate a hash list `value` against existing vectorhash binding

    :param binding: The existing `VectorHashBinding`
    :param value: the value to be validated (`HashList`)

    :return valid, invalid: (HashList) The values could contain [None, Hash()]
    """

    def _validate_row(row_bindings, row_hash):
        """Validate a single row of the table

        :param row_bindings: The `rowSchema` attribute binding
        :param row_hash: The hash to be validated, either `Hash` or `None`

        :returns:

            - valid: Either a `Hash` or `None`. `None` if validation failed.
            - invalid: Either a `Hash` or `None`
        """
        if row_hash is None:
            # If `row_hash` Hash is `None`, we report as invalid value
            return None, Hash()

        valid = Hash()
        # Check if order of the keys are respected...
        if list(row_bindings.keys()) != list(row_hash.keys()):
            row_hash = realign_hash(row_hash, reference=row_bindings.keys())

        for path, value in row_hash.items():
            binding = row_bindings.get(path, None)
            if binding is None:
                # Tables might lose a property in a row schema. This is fine
                # as we can simply ignore this case. We don't report this as
                # invalid but continue gracefully ...
                continue

            if value is None:
                # The property value is `None` or the value doesn't exist
                # (from `realign_hash`).
                # Try to use the default value from the binding!
                validated_value = get_default_value(binding)
            else:
                validated_value = validate_value(binding, value)

            if validated_value is None:
                # We report the invalid property
                return None, row_hash

            valid[path] = validated_value

        return valid, None

    # Set default values
    valid, invalid = HashList(), HashList()
    for row_hash in value:
        valid_row, invalid_row = _validate_row(binding.bindings, row_hash)
        if valid_row is not None:
            valid.append(valid_row)
        if invalid_row is not None:
            invalid.append(invalid_row)

    return valid, invalid


def get_default_value(binding, force=False):
    """Get the default value from a binding"""

    def _get_vector_default(binding):
        """ Provide a default value for vector binding `binding` considering
        minSize attribute"""
        attrs = binding.attributes
        min_size = attrs.get(const.KARABO_SCHEMA_MIN_SIZE)
        if min_size is None:
            return []

        if isinstance(binding, VectorNumberBinding):
            numpy = get_numpy_binding(binding)
            value = np.zeros(shape=(min_size,), dtype=numpy)
        elif isinstance(binding, VectorStringBinding):
            value = [""] * min_size
        else:
            value = []
        return value

    def _get_binding_default(binding):
        # Provide a default value for all leafType bindings of `binding`
        if isinstance(binding, (CharBinding, StringBinding)):
            return ""
        if isinstance(binding, IntBinding):
            return validate_binding_minimum(binding, 0)
        elif isinstance(binding, FloatBinding):
            return validate_binding_minimum(binding, 0.0)
        elif isinstance(binding, VectorBinding):
            # All vectors including table!
            return _get_vector_default(binding)
        elif isinstance(binding, BoolBinding):
            return False
        elif isinstance(binding, ByteArrayBinding):
            return bytearray([])

        return None

    attrs = binding.attributes
    value = attrs.get(const.KARABO_SCHEMA_DEFAULT_VALUE, None)
    if value is None and force:
        options = attrs.get(const.KARABO_SCHEMA_OPTIONS, None)
        if options is not None and len(options):
            value = options[0]
        else:
            value = _get_binding_default(binding)
            assert value is not None, f"No default value for {type(binding)}"

    return value


def convert_string(value: str):
    """Try to convert the string value `value` to a literal

    :returns: Tuple of value and success boolean
    """
    try:
        return literal_eval(value), True
    except (SyntaxError, ValueError):
        # Conversion of string to a literal failed, we return the value as is.
        return value, False


def validate_binding_minimum(binding, value):
    """Check the binding `binding` for a minimum value and align the
    value `value` if necessary

    The binding must be of an instance of `IntBinding` or `FloatBinding`
    """
    attrs = binding.attributes
    if isinstance(binding, IntBinding):
        range_trait = binding.trait('value').handler
        trait_low = range_trait._low
        if range_trait._exclude_low:
            trait_low = trait_low + 1

        low = attrs.get(const.KARABO_SCHEMA_MIN_EXC)
        if low is not None:
            low += 1
        else:
            low = attrs.get(const.KARABO_SCHEMA_MIN_INC, trait_low)

        if value >= low:
            return value

        return low

    elif isinstance(binding, FloatBinding):
        value_type = attrs.get(const.KARABO_SCHEMA_VALUE_TYPE)
        if value_type in ('FLOAT', 'COMPLEX_FLOAT'):
            info = np.finfo(np.float32)
        else:
            info = np.finfo(np.float64)

        low = attrs.get(const.KARABO_SCHEMA_MIN_EXC)
        if low is not None:
            low = low * (1 + np.sign(low) * info.eps) + info.tiny
        else:
            low = attrs.get(const.KARABO_SCHEMA_MIN_INC, info.min)

        if value >= low:
            return value

        return low

    return value


def validate_binding_configuration(binding, config):
    """Validate a configuration against a `binding` from a class schema"""
    assert isinstance(binding, BindingRoot)

    def _iter_binding(node, base=''):
        namespace = node.value
        base = base + '.' if base else ''
        for name in namespace:
            subname = base + name
            subnode = getattr(namespace, name)
            if isinstance(subnode, NodeBinding):
                yield from _iter_binding(subnode, base=subname)
            elif not isinstance(subnode, SlotBinding):
                yield subname, subnode

    fails = Hash()
    for key, node in _iter_binding(binding):
        if key in config:
            value = config[key]
            if isinstance(node, VectorHashBinding):
                if value is None:
                    # Validate table expects a HashList
                    fails.update({key: value})
                else:
                    _, invalid = validate_table_value(node, value)
                    if invalid:
                        fails.update({key: value})
            elif validate_value(node, value) is None:
                fails.update({key: value})

    return fails
